from typing import List, Optional, Sequence

from . import xdr as stellar_xdr
from .ledger_bounds import LedgerBounds
from .signer_key import SignerKey
from .time_bounds import TimeBounds


class Preconditions:
    """This contains a set of conditions, if a transaction can be accepted by the network,
    it must meet these conditions.

    :param time_bounds: The timebounds for the transaction.
    :param ledger_bounds: The ledgerbounds for the transaction.
    :param min_sequence_number: The minimum source account sequence
        number this transaction is valid for. If the value is ``None``,
        the transaction is valid when **source account's sequence number == tx.sequence - 1**.
    :param min_sequence_age: The minimum amount of time between
        source account sequence time and the ledger time when this transaction
        will become valid. If the value is ``0`` or ``None``, the transaction is unrestricted
        by the account sequence age. Cannot be negative.
    :param min_sequence_ledger_gap: The minimum number of ledgers between source account
        sequence and the ledger number when this transaction will become valid.
        If the value is ``0`` or ``None``, the transaction is unrestricted by the account sequence
        ledger. Cannot be negative.
    :param extra_signers: required extra signers.
    """

    def __init__(
        self,
        time_bounds: TimeBounds = None,
        ledger_bounds: LedgerBounds = None,
        min_sequence_number: int = None,
        min_sequence_age: int = None,
        min_sequence_ledger_gap: int = None,
        extra_signers: Sequence[SignerKey] = None,
    ):
        if not extra_signers:
            extra_signers = []

        if len(extra_signers) > 2:
            raise ValueError('"extra_signers" cannot be longer than 2 elements.')

        self.time_bounds = time_bounds
        self.ledger_bounds = ledger_bounds
        self.min_sequence_number = min_sequence_number
        self.min_sequence_age = min_sequence_age
        self.min_sequence_ledger_gap = min_sequence_ledger_gap
        self.extra_signers = extra_signers

        if self._is_v2():
            self.min_sequence_age = (
                0 if self.min_sequence_age is None else self.min_sequence_age
            )
            self.min_sequence_ledger_gap = (
                0
                if self.min_sequence_ledger_gap is None
                else self.min_sequence_ledger_gap
            )

    def _is_empty_preconditions(self) -> bool:
        return not (
            self.time_bounds
            or self.ledger_bounds
            or self.min_sequence_number is not None
            or self.min_sequence_age is not None
            or self.min_sequence_ledger_gap is not None
            or self.extra_signers
        )

    def _is_v2(self) -> bool:
        return bool(
            self.ledger_bounds
            or self.min_sequence_number is not None
            or self.min_sequence_age is not None
            or self.min_sequence_ledger_gap is not None
            or self.extra_signers
        )

    def to_xdr_object(self) -> stellar_xdr.Preconditions:
        """Returns the xdr object for this Preconditions object.

        :return: XDR Preconditions object
        """
        time_bounds = self.time_bounds.to_xdr_object() if self.time_bounds else None
        if self._is_v2():
            ledger_bounds = (
                self.ledger_bounds.to_xdr_object() if self.ledger_bounds else None
            )
            min_sequence_number = (
                stellar_xdr.SequenceNumber(stellar_xdr.Int64(self.min_sequence_number))
                if self.min_sequence_number is not None
                else None
            )
            min_sequence_age = (
                stellar_xdr.Duration(stellar_xdr.Uint64(self.min_sequence_age))
                if self.min_sequence_age is not None
                else stellar_xdr.Duration(stellar_xdr.Uint64(0))
            )
            min_sequence_ledger_gap = (
                stellar_xdr.Uint32(self.min_sequence_ledger_gap)
                if self.min_sequence_ledger_gap is not None
                else stellar_xdr.Uint32(0)
            )
            extra_signers = []
            if self.extra_signers:
                for s in self.extra_signers:
                    extra_signers.append(s.to_xdr_object())
            preconditions_v2 = stellar_xdr.PreconditionsV2(
                time_bounds=time_bounds,
                ledger_bounds=ledger_bounds,
                min_seq_num=min_sequence_number,
                min_seq_age=min_sequence_age,
                min_seq_ledger_gap=min_sequence_ledger_gap,
                extra_signers=extra_signers,
            )
            preconditions = stellar_xdr.Preconditions(
                stellar_xdr.PreconditionType.PRECOND_V2,
                v2=preconditions_v2,
            )
        elif time_bounds:
            preconditions = stellar_xdr.Preconditions(
                stellar_xdr.PreconditionType.PRECOND_TIME, time_bounds=time_bounds
            )
        else:
            preconditions = stellar_xdr.Preconditions(
                stellar_xdr.PreconditionType.PRECOND_NONE
            )
        return preconditions

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Preconditions) -> "Preconditions":
        """Create a :class:`Preconditions` from an XDR Preconditions object.

        :param xdr_object: The XDR LedgerBounds object.
        :return: A new :class:`LedgerBounds` object from the given XDR LedgerBounds object.
        """

        if xdr_object.type == stellar_xdr.PreconditionType.PRECOND_V2:
            assert xdr_object.v2 is not None
            time_bounds = (
                TimeBounds.from_xdr_object(xdr_object.v2.time_bounds)
                if xdr_object.v2.time_bounds
                else None
            )
            ledger_bounds = (
                LedgerBounds.from_xdr_object(xdr_object.v2.ledger_bounds)
                if xdr_object.v2.ledger_bounds
                else None
            )
            # min_sequence_number is nullable
            min_sequence_number = (
                xdr_object.v2.min_seq_num.sequence_number.int64
                if xdr_object.v2.min_seq_num is not None
                else None
            )
            min_sequence_age = xdr_object.v2.min_seq_age.duration.uint64
            min_sequence_ledger_gap = xdr_object.v2.min_seq_ledger_gap.uint32
            extra_signers: Optional[List[SignerKey]] = [
                SignerKey.from_xdr_object(s) for s in xdr_object.v2.extra_signers
            ]
            return cls(
                time_bounds=time_bounds,
                ledger_bounds=ledger_bounds,
                min_sequence_number=min_sequence_number,
                min_sequence_age=min_sequence_age,
                min_sequence_ledger_gap=min_sequence_ledger_gap,
                extra_signers=extra_signers,
            )
        elif xdr_object.type == stellar_xdr.PreconditionType.PRECOND_TIME:
            time_bounds = (
                TimeBounds.from_xdr_object(xdr_object.time_bounds)
                if xdr_object.time_bounds
                else None
            )
            return cls(time_bounds=time_bounds)
        elif xdr_object.type == stellar_xdr.PreconditionType.PRECOND_NONE:
            return cls()
        else:
            raise ValueError(f"Invalid PreconditionType: {xdr_object.type!r}")

    def __hash__(self):
        return hash(
            (
                self.time_bounds,
                self.ledger_bounds,
                self.min_sequence_number,
                self.min_sequence_age,
                self.min_sequence_ledger_gap,
                self.extra_signers,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.time_bounds == other.time_bounds
            and self.ledger_bounds == other.ledger_bounds
            and self.min_sequence_number == other.min_sequence_number
            and self.min_sequence_age == other.min_sequence_age
            and self.min_sequence_ledger_gap == other.min_sequence_ledger_gap
            and self.extra_signers == other.extra_signers
        )

    def __repr__(self):
        return (
            f"<Preconditions ["
            f"time_bounds={self.time_bounds}, "
            f"ledger_bounds={self.ledger_bounds}, "
            f"min_sequence_number={self.min_sequence_number}, "
            f"min_sequence_age={self.min_sequence_age}, "
            f"min_sequence_ledger_gap={self.min_sequence_ledger_gap}, "
            f"extra_signers={self.extra_signers}"
            f"]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Integer
from .generalized_transaction_set import GeneralizedTransactionSet
from .transaction_set import TransactionSet

__all__ = ["StoredTransactionSet"]


class StoredTransactionSet:
    """
    XDR Source Code::

                                                                union StoredTransactionSet switch (int v)
                                                                {
                                                                case 0:
                                                                        TransactionSet txSet;
                                                                case 1:
                                                                        GeneralizedTransactionSet generalizedTxSet;
                                                                };
    """

    def __init__(
        self,
        v: int,
        tx_set: TransactionSet = None,
        generalized_tx_set: GeneralizedTransactionSet = None,
    ) -> None:
        self.v = v
        self.tx_set = tx_set
        self.generalized_tx_set = generalized_tx_set

    def pack(self, packer: Packer) -> None:
        Integer(self.v).pack(packer)
        if self.v == 0:
            if self.tx_set is None:
                raise ValueError("tx_set should not be None.")
            self.tx_set.pack(packer)
            return
        if self.v == 1:
            if self.generalized_tx_set is None:
                raise ValueError("generalized_tx_set should not be None.")
            self.generalized_tx_set.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> StoredTransactionSet:
        v = Integer.unpack(unpacker)
        if v == 0:
            tx_set = TransactionSet.unpack(unpacker)
            return cls(v=v, tx_set=tx_set)
        if v == 1:
            generalized_tx_set = GeneralizedTransactionSet.unpack(unpacker)
            return cls(v=v, generalized_tx_set=generalized_tx_set)
        return cls(v=v)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> StoredTransactionSet:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> StoredTransactionSet:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.v,
                self.tx_set,
                self.generalized_tx_set,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.v == other.v
            and self.tx_set == other.tx_set
            and self.generalized_tx_set == other.generalized_tx_set
        )

    def __repr__(self):
        out = []
        out.append(f"v={self.v}")
        out.append(f"tx_set={self.tx_set}") if self.tx_set is not None else None
        (
            out.append(f"generalized_tx_set={self.generalized_tx_set}")
            if self.generalized_tx_set is not None
            else None
        )
        return f"<StoredTransactionSet [{', '.join(out)}]>"



from typing import Sequence, Union

from xdrlib3 import Packer

from . import xdr as stellar_xdr
from .base_transaction_envelope import BaseTransactionEnvelope
from .decorated_signature import DecoratedSignature
from .fee_bump_transaction import FeeBumpTransaction

__all__ = ["FeeBumpTransactionEnvelope"]


class FeeBumpTransactionEnvelope(BaseTransactionEnvelope["FeeBumpTransactionEnvelope"]):
    """The :class:`FeeBumpTransactionEnvelope` object, which represents a fee bump transaction
    envelope ready to sign and submit to send over the network.

    When a fee bump transaction is ready to be prepared for sending over the network, it
    must be put into a :class:`FeeBumpTransactionEnvelope`, which includes additional
    metadata such as the signers for a given transaction. Ultimately, this
    class handles signing and conversion to and from XDR for usage on Stellar's
    network.

    See `Fee-Bump Transactions <https://developers.stellar.org/docs/glossary/fee-bumps/>`__ for more information.
    See `CAP-0015 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0015.md>`__ for more information.

    :param transaction: The fee bump transaction that is encapsulated in this envelope.
    :param signatures: which contains a list of signatures that have
          already been created.
    :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
    """

    def __init__(
        self,
        transaction: FeeBumpTransaction,
        network_passphrase: str,
        signatures: Sequence[DecoratedSignature] = None,
    ) -> None:
        super().__init__(network_passphrase, signatures)
        self.transaction: FeeBumpTransaction = transaction

    def signature_base(self) -> bytes:
        """Get the signature base of this transaction envelope.

        Return the "signature base" of this transaction, which is the value
        that, when hashed, should be signed to create a signature that
        validators on the Stellar Network will accept.

        It is composed of a 4 prefix bytes followed by the xdr-encoded form of
        this transaction.

        :return: The signature base of this transaction envelope.

        """
        network_id = self._network_id
        packer = Packer()
        stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP.pack(packer)
        self.transaction.to_xdr_object().pack(packer)
        return network_id + packer.get_buffer()

    @staticmethod
    def is_fee_bump_transaction_envelope(xdr: Union[str, bytes]) -> bool:
        if isinstance(xdr, str):
            xdr_object = stellar_xdr.TransactionEnvelope.from_xdr(xdr)
        else:
            xdr_object = stellar_xdr.TransactionEnvelope.from_xdr_bytes(xdr)
        te_type = xdr_object.type
        if te_type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP:
            return True
        elif (
            te_type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX
            or te_type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_V0
        ):
            return False
        else:
            raise ValueError(
                f"This transaction envelope type is not supported, type = {te_type}."
            )

    def to_xdr_object(self) -> stellar_xdr.TransactionEnvelope:
        """Get an XDR object representation of this :class:`TransactionEnvelope`.

        :return: XDR TransactionEnvelope object
        """
        tx = self.transaction.to_xdr_object()
        te_type = stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP
        signatures = [signature.to_xdr_object() for signature in self.signatures]
        tx_envelope = stellar_xdr.FeeBumpTransactionEnvelope(tx, signatures)
        return stellar_xdr.TransactionEnvelope(type=te_type, fee_bump=tx_envelope)

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.TransactionEnvelope, network_passphrase: str
    ) -> "FeeBumpTransactionEnvelope":
        """Create a new :class:`FeeBumpTransactionEnvelope` from an XDR object.

        :param xdr_object: The XDR object that represents a fee bump transaction envelope.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        :return: A new :class:`FeeBumpTransactionEnvelope` object from the given XDR TransactionEnvelope object.
        """
        te_type = xdr_object.type
        if te_type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP:
            assert xdr_object.fee_bump is not None
            tx = FeeBumpTransaction.from_xdr_object(
                xdr_object.fee_bump.tx, network_passphrase
            )
        else:
            raise ValueError("Invalid EnvelopeType: %d.", xdr_object.type)
        assert xdr_object.fee_bump is not None
        signatures = [
            DecoratedSignature.from_xdr_object(s)
            for s in xdr_object.fee_bump.signatures
        ]
        te = cls(tx, network_passphrase=network_passphrase, signatures=signatures)
        return te

    def __hash__(self):
        return hash((self.transaction, self.network_passphrase, self.signatures))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.transaction == other.transaction
            and self.network_passphrase == other.network_passphrase
            and self.signatures == other.signatures
        )

    def __repr__(self):
        return (
            f"<FeeBumpTransactionEnvelope [transaction={self.transaction}, "
            f"network_passphrase={self.network_passphrase}, signatures={self.signatures}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["ConfigSettingID"]


class ConfigSettingID(IntEnum):
    """
    XDR Source Code::

        enum ConfigSettingID
        {
            CONFIG_SETTING_CONTRACT_MAX_SIZE_BYTES = 0,
            CONFIG_SETTING_CONTRACT_COMPUTE_V0 = 1,
            CONFIG_SETTING_CONTRACT_LEDGER_COST_V0 = 2,
            CONFIG_SETTING_CONTRACT_HISTORICAL_DATA_V0 = 3,
            CONFIG_SETTING_CONTRACT_EVENTS_V0 = 4,
            CONFIG_SETTING_CONTRACT_BANDWIDTH_V0 = 5,
            CONFIG_SETTING_CONTRACT_COST_PARAMS_CPU_INSTRUCTIONS = 6,
            CONFIG_SETTING_CONTRACT_COST_PARAMS_MEMORY_BYTES = 7,
            CONFIG_SETTING_CONTRACT_DATA_KEY_SIZE_BYTES = 8,
            CONFIG_SETTING_CONTRACT_DATA_ENTRY_SIZE_BYTES = 9,
            CONFIG_SETTING_STATE_ARCHIVAL = 10,
            CONFIG_SETTING_CONTRACT_EXECUTION_LANES = 11,
            CONFIG_SETTING_BUCKETLIST_SIZE_WINDOW = 12,
            CONFIG_SETTING_EVICTION_ITERATOR = 13
        };
    """

    CONFIG_SETTING_CONTRACT_MAX_SIZE_BYTES = 0
    CONFIG_SETTING_CONTRACT_COMPUTE_V0 = 1
    CONFIG_SETTING_CONTRACT_LEDGER_COST_V0 = 2
    CONFIG_SETTING_CONTRACT_HISTORICAL_DATA_V0 = 3
    CONFIG_SETTING_CONTRACT_EVENTS_V0 = 4
    CONFIG_SETTING_CONTRACT_BANDWIDTH_V0 = 5
    CONFIG_SETTING_CONTRACT_COST_PARAMS_CPU_INSTRUCTIONS = 6
    CONFIG_SETTING_CONTRACT_COST_PARAMS_MEMORY_BYTES = 7
    CONFIG_SETTING_CONTRACT_DATA_KEY_SIZE_BYTES = 8
    CONFIG_SETTING_CONTRACT_DATA_ENTRY_SIZE_BYTES = 9
    CONFIG_SETTING_STATE_ARCHIVAL = 10
    CONFIG_SETTING_CONTRACT_EXECUTION_LANES = 11
    CONFIG_SETTING_BUCKETLIST_SIZE_WINDOW = 12
    CONFIG_SETTING_EVICTION_ITERATOR = 13

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigSettingID:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigSettingID:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigSettingID:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Integer

__all__ = ["ClaimableBalanceEntryExtensionV1Ext"]


class ClaimableBalanceEntryExtensionV1Ext:
    """
    XDR Source Code::

        union switch (int v)
            {
            case 0:
                void;
            }
    """

    def __init__(
        self,
        v: int,
    ) -> None:
        self.v = v

    def pack(self, packer: Packer) -> None:
        Integer(self.v).pack(packer)
        if self.v == 0:
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClaimableBalanceEntryExtensionV1Ext:
        v = Integer.unpack(unpacker)
        if v == 0:
            return cls(v=v)
        return cls(v=v)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClaimableBalanceEntryExtensionV1Ext:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClaimableBalanceEntryExtensionV1Ext:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash((self.v,))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.v == other.v

    def __repr__(self):
        out = []
        out.append(f"v={self.v}")
        return f"<ClaimableBalanceEntryExtensionV1Ext [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["TrustLineFlags"]


class TrustLineFlags(IntEnum):
    """
    XDR Source Code::

        enum TrustLineFlags
        {
            // issuer has authorized account to perform transactions with its credit
            AUTHORIZED_FLAG = 1,
            // issuer has authorized account to maintain and reduce liabilities for its
            // credit
            AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG = 2,
            // issuer has specified that it may clawback its credit, and that claimable
            // balances created with its credit may also be clawed back
            TRUSTLINE_CLAWBACK_ENABLED_FLAG = 4
        };
    """

    AUTHORIZED_FLAG = 1
    AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG = 2
    TRUSTLINE_CLAWBACK_ENABLED_FLAG = 4

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TrustLineFlags:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TrustLineFlags:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TrustLineFlags:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from typing import Optional

from . import xdr as stellar_xdr
from .keypair import Keypair
from .strkey import StrKey

__all__ = ["MuxedAccount"]


class MuxedAccount:
    """The :class:`MuxedAccount` object, which represents a multiplexed account on Stellar's network.

    An example::

        from stellar_sdk import MuxedAccount

        account_id = "GAQAA5L65LSYH7CQ3VTJ7F3HHLGCL3DSLAR2Y47263D56MNNGHSQSTVY"
        account_muxed_id = 1234
        account_muxed = "MAQAA5L65LSYH7CQ3VTJ7F3HHLGCL3DSLAR2Y47263D56MNNGHSQSAAAAAAAAAAE2LP26"

        # generate account_muxed
        muxed = MuxedAccount(account=account_id, account_muxed_id=1234)  # account_muxed_id is optional.
        print(f"account_muxed: {muxed.account_muxed}")  # `account_muxed` returns ``None`` if `account_muxed_id` is ``None``.

        # parse account_muxed
        muxed = MuxedAccount.from_account(account_muxed)
        print(f"account_id: {muxed.account_id}\\n"
              f"account_muxed_id: {muxed.account_muxed_id}")


    See `SEP-0023 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0023.md>`_ for more
    information.

    :param account_id: ed25519 account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``.
        It should be a string starting with ``G``. If you want to build a MuxedAccount
        object using an address starting with ``M``, please use :func:`stellar_sdk.MuxedAccount.from_account`.
    :param account_muxed_id: account multiplexing id (ex. ``1234``)
    """

    def __init__(self, account_id: str, account_muxed_id: Optional[int] = None) -> None:
        Keypair.from_public_key(account_id)
        self.account_id: str = account_id
        self.account_muxed_id: Optional[int] = account_muxed_id

    @property
    def account_muxed(self) -> Optional[str]:
        """Get the multiplex address starting with ``M``, return ``None`` if `account_id_id` is ``None``."""

        if self.account_muxed_id is None:
            return None

        muxed_xdr = stellar_xdr.MuxedAccount(
            type=stellar_xdr.CryptoKeyType.KEY_TYPE_MUXED_ED25519,
            med25519=stellar_xdr.MuxedAccountMed25519(
                id=stellar_xdr.Uint64(self.account_muxed_id),
                ed25519=stellar_xdr.Uint256(
                    StrKey.decode_ed25519_public_key(self.account_id)
                ),
            ),
        )
        return StrKey.encode_muxed_account(muxed_xdr)

    @account_muxed.setter
    def account_muxed(self, value):
        raise AttributeError(
            "Can't set attribute, use `MuxedAccount.from_account` instead."
        )

    @property
    def universal_account_id(self) -> str:
        """Get the universal account id,
        if `account_muxed_id` is ``None``, it will return ed25519
        public key (ex. ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``),
        otherwise it will return muxed
        account (ex. ``"MAAAAAAAAAAAJURAAB2X52XFQP6FBXLGT6LWOOWMEXWHEWBDVRZ7V5WH34Y22MPFBHUHY"``)
        """
        if self.account_muxed_id is None:
            return self.account_id
        assert self.account_muxed is not None
        return self.account_muxed

    @classmethod
    def from_account(cls, account: str) -> "MuxedAccount":
        """Create a :class:`MuxedAccount` object from account id or muxed account id.

        :param account: account id
            or muxed account id (ex. ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
            or ``"MAAAAAAAAAAAJURAAB2X52XFQP6FBXLGT6LWOOWMEXWHEWBDVRZ7V5WH34Y22MPFBHUHY"``)
        """
        data_length = len(account)
        if data_length == 56:
            return cls(account_id=account, account_muxed_id=None)
        elif data_length == 69:
            muxed_xdr = StrKey.decode_muxed_account(account)
            assert muxed_xdr.med25519 is not None
            assert muxed_xdr.med25519.ed25519 is not None
            return cls(
                account_id=StrKey.encode_ed25519_public_key(
                    muxed_xdr.med25519.ed25519.uint256
                ),
                account_muxed_id=muxed_xdr.med25519.id.uint64,
            )
        else:
            raise ValueError(f"This is not a valid account: {account}")

    def to_xdr_object(self) -> stellar_xdr.MuxedAccount:
        """Returns the xdr object for this MuxedAccount object.

        :return: XDR MuxedAccount object
        """
        if self.account_muxed_id is None:
            return StrKey.decode_muxed_account(self.account_id)
        assert self.account_muxed is not None
        return StrKey.decode_muxed_account(self.account_muxed)

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.MuxedAccount) -> "MuxedAccount":
        """Create a :class:`MuxedAccount` object from an XDR Asset object.

        :param xdr_object: The MuxedAccount object.
        :return: A new :class:`MuxedAccount` object from the given XDR MuxedAccount object.
        """
        if xdr_object.type == stellar_xdr.CryptoKeyType.KEY_TYPE_ED25519:
            assert xdr_object.ed25519 is not None
            account_id = StrKey.encode_ed25519_public_key(xdr_object.ed25519.uint256)
            return cls(account_id=account_id, account_muxed_id=None)
        assert xdr_object.med25519 is not None
        account_id_id = xdr_object.med25519.id.uint64
        assert xdr_object.med25519 is not None
        account_id = StrKey.encode_ed25519_public_key(
            xdr_object.med25519.ed25519.uint256
        )
        return cls(account_id=account_id, account_muxed_id=account_id_id)

    def __hash__(self):
        return hash((self.account_id, self.account_muxed_id))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.account_id == other.account_id
            and self.account_muxed_id == other.account_muxed_id
        )

    def __repr__(self):
        return "<MuxedAccount [account_id={account_id}, account_muxed_id={account_muxed_id}]>".format(
            account_id=self.account_id, account_muxed_id=self.account_muxed_id
        )



from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseRootCallBuilder"]


class BaseRootCallBuilder(BaseCallBuilder):
    """Creates a new :class:`RootCallBuilder` pointed to server defined by horizon_url.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .auth import Auth
from .dont_have import DontHave
from .error import Error
from .flood_advert import FloodAdvert
from .flood_demand import FloodDemand
from .generalized_transaction_set import GeneralizedTransactionSet
from .hello import Hello
from .message_type import MessageType
from .peer_address import PeerAddress
from .scp_envelope import SCPEnvelope
from .scp_quorum_set import SCPQuorumSet
from .send_more import SendMore
from .send_more_extended import SendMoreExtended
from .signed_survey_request_message import SignedSurveyRequestMessage
from .signed_survey_response_message import SignedSurveyResponseMessage
from .signed_time_sliced_survey_request_message import (
    SignedTimeSlicedSurveyRequestMessage,
)
from .signed_time_sliced_survey_response_message import (
    SignedTimeSlicedSurveyResponseMessage,
)
from .signed_time_sliced_survey_start_collecting_message import (
    SignedTimeSlicedSurveyStartCollectingMessage,
)
from .signed_time_sliced_survey_stop_collecting_message import (
    SignedTimeSlicedSurveyStopCollectingMessage,
)
from .transaction_envelope import TransactionEnvelope
from .transaction_set import TransactionSet
from .uint32 import Uint32
from .uint256 import Uint256

__all__ = ["StellarMessage"]


class StellarMessage:
    """
    XDR Source Code::

        union StellarMessage switch (MessageType type)
        {
        case ERROR_MSG:
            Error error;
        case HELLO:
            Hello hello;
        case AUTH:
            Auth auth;
        case DONT_HAVE:
            DontHave dontHave;
        case GET_PEERS:
            void;
        case PEERS:
            PeerAddress peers<100>;

        case GET_TX_SET:
            uint256 txSetHash;
        case TX_SET:
            TransactionSet txSet;
        case GENERALIZED_TX_SET:
            GeneralizedTransactionSet generalizedTxSet;

        case TRANSACTION:
            TransactionEnvelope transaction;

        case SURVEY_REQUEST:
            SignedSurveyRequestMessage signedSurveyRequestMessage;

        case SURVEY_RESPONSE:
            SignedSurveyResponseMessage signedSurveyResponseMessage;

        case TIME_SLICED_SURVEY_REQUEST:
            SignedTimeSlicedSurveyRequestMessage signedTimeSlicedSurveyRequestMessage;

        case TIME_SLICED_SURVEY_RESPONSE:
            SignedTimeSlicedSurveyResponseMessage signedTimeSlicedSurveyResponseMessage;

        case TIME_SLICED_SURVEY_START_COLLECTING:
            SignedTimeSlicedSurveyStartCollectingMessage
                signedTimeSlicedSurveyStartCollectingMessage;

        case TIME_SLICED_SURVEY_STOP_COLLECTING:
            SignedTimeSlicedSurveyStopCollectingMessage
                signedTimeSlicedSurveyStopCollectingMessage;

        // SCP
        case GET_SCP_QUORUMSET:
            uint256 qSetHash;
        case SCP_QUORUMSET:
            SCPQuorumSet qSet;
        case SCP_MESSAGE:
            SCPEnvelope envelope;
        case GET_SCP_STATE:
            uint32 getSCPLedgerSeq; // ledger seq requested ; if 0, requests the latest
        case SEND_MORE:
            SendMore sendMoreMessage;
        case SEND_MORE_EXTENDED:
            SendMoreExtended sendMoreExtendedMessage;
        // Pull mode
        case FLOOD_ADVERT:
             FloodAdvert floodAdvert;
        case FLOOD_DEMAND:
             FloodDemand floodDemand;
        };
    """

    def __init__(
        self,
        type: MessageType,
        error: Error = None,
        hello: Hello = None,
        auth: Auth = None,
        dont_have: DontHave = None,
        peers: List[PeerAddress] = None,
        tx_set_hash: Uint256 = None,
        tx_set: TransactionSet = None,
        generalized_tx_set: GeneralizedTransactionSet = None,
        transaction: TransactionEnvelope = None,
        signed_survey_request_message: SignedSurveyRequestMessage = None,
        signed_survey_response_message: SignedSurveyResponseMessage = None,
        signed_time_sliced_survey_request_message: SignedTimeSlicedSurveyRequestMessage = None,
        signed_time_sliced_survey_response_message: SignedTimeSlicedSurveyResponseMessage = None,
        signed_time_sliced_survey_start_collecting_message: SignedTimeSlicedSurveyStartCollectingMessage = None,
        signed_time_sliced_survey_stop_collecting_message: SignedTimeSlicedSurveyStopCollectingMessage = None,
        q_set_hash: Uint256 = None,
        q_set: SCPQuorumSet = None,
        envelope: SCPEnvelope = None,
        get_scp_ledger_seq: Uint32 = None,
        send_more_message: SendMore = None,
        send_more_extended_message: SendMoreExtended = None,
        flood_advert: FloodAdvert = None,
        flood_demand: FloodDemand = None,
    ) -> None:
        _expect_max_length = 100
        if peers and len(peers) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `peers` should be {_expect_max_length}, but got {len(peers)}."
            )
        self.type = type
        self.error = error
        self.hello = hello
        self.auth = auth
        self.dont_have = dont_have
        self.peers = peers
        self.tx_set_hash = tx_set_hash
        self.tx_set = tx_set
        self.generalized_tx_set = generalized_tx_set
        self.transaction = transaction
        self.signed_survey_request_message = signed_survey_request_message
        self.signed_survey_response_message = signed_survey_response_message
        self.signed_time_sliced_survey_request_message = (
            signed_time_sliced_survey_request_message
        )
        self.signed_time_sliced_survey_response_message = (
            signed_time_sliced_survey_response_message
        )
        self.signed_time_sliced_survey_start_collecting_message = (
            signed_time_sliced_survey_start_collecting_message
        )
        self.signed_time_sliced_survey_stop_collecting_message = (
            signed_time_sliced_survey_stop_collecting_message
        )
        self.q_set_hash = q_set_hash
        self.q_set = q_set
        self.envelope = envelope
        self.get_scp_ledger_seq = get_scp_ledger_seq
        self.send_more_message = send_more_message
        self.send_more_extended_message = send_more_extended_message
        self.flood_advert = flood_advert
        self.flood_demand = flood_demand

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == MessageType.ERROR_MSG:
            if self.error is None:
                raise ValueError("error should not be None.")
            self.error.pack(packer)
            return
        if self.type == MessageType.HELLO:
            if self.hello is None:
                raise ValueError("hello should not be None.")
            self.hello.pack(packer)
            return
        if self.type == MessageType.AUTH:
            if self.auth is None:
                raise ValueError("auth should not be None.")
            self.auth.pack(packer)
            return
        if self.type == MessageType.DONT_HAVE:
            if self.dont_have is None:
                raise ValueError("dont_have should not be None.")
            self.dont_have.pack(packer)
            return
        if self.type == MessageType.GET_PEERS:
            return
        if self.type == MessageType.PEERS:
            if self.peers is None:
                raise ValueError("peers should not be None.")
            packer.pack_uint(len(self.peers))
            for peers_item in self.peers:
                peers_item.pack(packer)
            return
        if self.type == MessageType.GET_TX_SET:
            if self.tx_set_hash is None:
                raise ValueError("tx_set_hash should not be None.")
            self.tx_set_hash.pack(packer)
            return
        if self.type == MessageType.TX_SET:
            if self.tx_set is None:
                raise ValueError("tx_set should not be None.")
            self.tx_set.pack(packer)
            return
        if self.type == MessageType.GENERALIZED_TX_SET:
            if self.generalized_tx_set is None:
                raise ValueError("generalized_tx_set should not be None.")
            self.generalized_tx_set.pack(packer)
            return
        if self.type == MessageType.TRANSACTION:
            if self.transaction is None:
                raise ValueError("transaction should not be None.")
            self.transaction.pack(packer)
            return
        if self.type == MessageType.SURVEY_REQUEST:
            if self.signed_survey_request_message is None:
                raise ValueError("signed_survey_request_message should not be None.")
            self.signed_survey_request_message.pack(packer)
            return
        if self.type == MessageType.SURVEY_RESPONSE:
            if self.signed_survey_response_message is None:
                raise ValueError("signed_survey_response_message should not be None.")
            self.signed_survey_response_message.pack(packer)
            return
        if self.type == MessageType.TIME_SLICED_SURVEY_REQUEST:
            if self.signed_time_sliced_survey_request_message is None:
                raise ValueError(
                    "signed_time_sliced_survey_request_message should not be None."
                )
            self.signed_time_sliced_survey_request_message.pack(packer)
            return
        if self.type == MessageType.TIME_SLICED_SURVEY_RESPONSE:
            if self.signed_time_sliced_survey_response_message is None:
                raise ValueError(
                    "signed_time_sliced_survey_response_message should not be None."
                )
            self.signed_time_sliced_survey_response_message.pack(packer)
            return
        if self.type == MessageType.TIME_SLICED_SURVEY_START_COLLECTING:
            if self.signed_time_sliced_survey_start_collecting_message is None:
                raise ValueError(
                    "signed_time_sliced_survey_start_collecting_message should not be None."
                )
            self.signed_time_sliced_survey_start_collecting_message.pack(packer)
            return
        if self.type == MessageType.TIME_SLICED_SURVEY_STOP_COLLECTING:
            if self.signed_time_sliced_survey_stop_collecting_message is None:
                raise ValueError(
                    "signed_time_sliced_survey_stop_collecting_message should not be None."
                )
            self.signed_time_sliced_survey_stop_collecting_message.pack(packer)
            return
        if self.type == MessageType.GET_SCP_QUORUMSET:
            if self.q_set_hash is None:
                raise ValueError("q_set_hash should not be None.")
            self.q_set_hash.pack(packer)
            return
        if self.type == MessageType.SCP_QUORUMSET:
            if self.q_set is None:
                raise ValueError("q_set should not be None.")
            self.q_set.pack(packer)
            return
        if self.type == MessageType.SCP_MESSAGE:
            if self.envelope is None:
                raise ValueError("envelope should not be None.")
            self.envelope.pack(packer)
            return
        if self.type == MessageType.GET_SCP_STATE:
            if self.get_scp_ledger_seq is None:
                raise ValueError("get_scp_ledger_seq should not be None.")
            self.get_scp_ledger_seq.pack(packer)
            return
        if self.type == MessageType.SEND_MORE:
            if self.send_more_message is None:
                raise ValueError("send_more_message should not be None.")
            self.send_more_message.pack(packer)
            return
        if self.type == MessageType.SEND_MORE_EXTENDED:
            if self.send_more_extended_message is None:
                raise ValueError("send_more_extended_message should not be None.")
            self.send_more_extended_message.pack(packer)
            return
        if self.type == MessageType.FLOOD_ADVERT:
            if self.flood_advert is None:
                raise ValueError("flood_advert should not be None.")
            self.flood_advert.pack(packer)
            return
        if self.type == MessageType.FLOOD_DEMAND:
            if self.flood_demand is None:
                raise ValueError("flood_demand should not be None.")
            self.flood_demand.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> StellarMessage:
        type = MessageType.unpack(unpacker)
        if type == MessageType.ERROR_MSG:
            error = Error.unpack(unpacker)
            return cls(type=type, error=error)
        if type == MessageType.HELLO:
            hello = Hello.unpack(unpacker)
            return cls(type=type, hello=hello)
        if type == MessageType.AUTH:
            auth = Auth.unpack(unpacker)
            return cls(type=type, auth=auth)
        if type == MessageType.DONT_HAVE:
            dont_have = DontHave.unpack(unpacker)
            return cls(type=type, dont_have=dont_have)
        if type == MessageType.GET_PEERS:
            return cls(type=type)
        if type == MessageType.PEERS:
            length = unpacker.unpack_uint()
            peers = []
            for _ in range(length):
                peers.append(PeerAddress.unpack(unpacker))
            return cls(type=type, peers=peers)
        if type == MessageType.GET_TX_SET:
            tx_set_hash = Uint256.unpack(unpacker)
            return cls(type=type, tx_set_hash=tx_set_hash)
        if type == MessageType.TX_SET:
            tx_set = TransactionSet.unpack(unpacker)
            return cls(type=type, tx_set=tx_set)
        if type == MessageType.GENERALIZED_TX_SET:
            generalized_tx_set = GeneralizedTransactionSet.unpack(unpacker)
            return cls(type=type, generalized_tx_set=generalized_tx_set)
        if type == MessageType.TRANSACTION:
            transaction = TransactionEnvelope.unpack(unpacker)
            return cls(type=type, transaction=transaction)
        if type == MessageType.SURVEY_REQUEST:
            signed_survey_request_message = SignedSurveyRequestMessage.unpack(unpacker)
            return cls(
                type=type, signed_survey_request_message=signed_survey_request_message
            )
        if type == MessageType.SURVEY_RESPONSE:
            signed_survey_response_message = SignedSurveyResponseMessage.unpack(
                unpacker
            )
            return cls(
                type=type, signed_survey_response_message=signed_survey_response_message
            )
        if type == MessageType.TIME_SLICED_SURVEY_REQUEST:
            signed_time_sliced_survey_request_message = (
                SignedTimeSlicedSurveyRequestMessage.unpack(unpacker)
            )
            return cls(
                type=type,
                signed_time_sliced_survey_request_message=signed_time_sliced_survey_request_message,
            )
        if type == MessageType.TIME_SLICED_SURVEY_RESPONSE:
            signed_time_sliced_survey_response_message = (
                SignedTimeSlicedSurveyResponseMessage.unpack(unpacker)
            )
            return cls(
                type=type,
                signed_time_sliced_survey_response_message=signed_time_sliced_survey_response_message,
            )
        if type == MessageType.TIME_SLICED_SURVEY_START_COLLECTING:
            signed_time_sliced_survey_start_collecting_message = (
                SignedTimeSlicedSurveyStartCollectingMessage.unpack(unpacker)
            )
            return cls(
                type=type,
                signed_time_sliced_survey_start_collecting_message=signed_time_sliced_survey_start_collecting_message,
            )
        if type == MessageType.TIME_SLICED_SURVEY_STOP_COLLECTING:
            signed_time_sliced_survey_stop_collecting_message = (
                SignedTimeSlicedSurveyStopCollectingMessage.unpack(unpacker)
            )
            return cls(
                type=type,
                signed_time_sliced_survey_stop_collecting_message=signed_time_sliced_survey_stop_collecting_message,
            )
        if type == MessageType.GET_SCP_QUORUMSET:
            q_set_hash = Uint256.unpack(unpacker)
            return cls(type=type, q_set_hash=q_set_hash)
        if type == MessageType.SCP_QUORUMSET:
            q_set = SCPQuorumSet.unpack(unpacker)
            return cls(type=type, q_set=q_set)
        if type == MessageType.SCP_MESSAGE:
            envelope = SCPEnvelope.unpack(unpacker)
            return cls(type=type, envelope=envelope)
        if type == MessageType.GET_SCP_STATE:
            get_scp_ledger_seq = Uint32.unpack(unpacker)
            return cls(type=type, get_scp_ledger_seq=get_scp_ledger_seq)
        if type == MessageType.SEND_MORE:
            send_more_message = SendMore.unpack(unpacker)
            return cls(type=type, send_more_message=send_more_message)
        if type == MessageType.SEND_MORE_EXTENDED:
            send_more_extended_message = SendMoreExtended.unpack(unpacker)
            return cls(type=type, send_more_extended_message=send_more_extended_message)
        if type == MessageType.FLOOD_ADVERT:
            flood_advert = FloodAdvert.unpack(unpacker)
            return cls(type=type, flood_advert=flood_advert)
        if type == MessageType.FLOOD_DEMAND:
            flood_demand = FloodDemand.unpack(unpacker)
            return cls(type=type, flood_demand=flood_demand)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> StellarMessage:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> StellarMessage:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.error,
                self.hello,
                self.auth,
                self.dont_have,
                self.peers,
                self.tx_set_hash,
                self.tx_set,
                self.generalized_tx_set,
                self.transaction,
                self.signed_survey_request_message,
                self.signed_survey_response_message,
                self.signed_time_sliced_survey_request_message,
                self.signed_time_sliced_survey_response_message,
                self.signed_time_sliced_survey_start_collecting_message,
                self.signed_time_sliced_survey_stop_collecting_message,
                self.q_set_hash,
                self.q_set,
                self.envelope,
                self.get_scp_ledger_seq,
                self.send_more_message,
                self.send_more_extended_message,
                self.flood_advert,
                self.flood_demand,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.error == other.error
            and self.hello == other.hello
            and self.auth == other.auth
            and self.dont_have == other.dont_have
            and self.peers == other.peers
            and self.tx_set_hash == other.tx_set_hash
            and self.tx_set == other.tx_set
            and self.generalized_tx_set == other.generalized_tx_set
            and self.transaction == other.transaction
            and self.signed_survey_request_message
            == other.signed_survey_request_message
            and self.signed_survey_response_message
            == other.signed_survey_response_message
            and self.signed_time_sliced_survey_request_message
            == other.signed_time_sliced_survey_request_message
            and self.signed_time_sliced_survey_response_message
            == other.signed_time_sliced_survey_response_message
            and self.signed_time_sliced_survey_start_collecting_message
            == other.signed_time_sliced_survey_start_collecting_message
            and self.signed_time_sliced_survey_stop_collecting_message
            == other.signed_time_sliced_survey_stop_collecting_message
            and self.q_set_hash == other.q_set_hash
            and self.q_set == other.q_set
            and self.envelope == other.envelope
            and self.get_scp_ledger_seq == other.get_scp_ledger_seq
            and self.send_more_message == other.send_more_message
            and self.send_more_extended_message == other.send_more_extended_message
            and self.flood_advert == other.flood_advert
            and self.flood_demand == other.flood_demand
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"error={self.error}") if self.error is not None else None
        out.append(f"hello={self.hello}") if self.hello is not None else None
        out.append(f"auth={self.auth}") if self.auth is not None else None
        (
            out.append(f"dont_have={self.dont_have}")
            if self.dont_have is not None
            else None
        )
        out.append(f"peers={self.peers}") if self.peers is not None else None
        (
            out.append(f"tx_set_hash={self.tx_set_hash}")
            if self.tx_set_hash is not None
            else None
        )
        out.append(f"tx_set={self.tx_set}") if self.tx_set is not None else None
        (
            out.append(f"generalized_tx_set={self.generalized_tx_set}")
            if self.generalized_tx_set is not None
            else None
        )
        (
            out.append(f"transaction={self.transaction}")
            if self.transaction is not None
            else None
        )
        (
            out.append(
                f"signed_survey_request_message={self.signed_survey_request_message}"
            )
            if self.signed_survey_request_message is not None
            else None
        )
        (
            out.append(
                f"signed_survey_response_message={self.signed_survey_response_message}"
            )
            if self.signed_survey_response_message is not None
            else None
        )
        (
            out.append(
                f"signed_time_sliced_survey_request_message={self.signed_time_sliced_survey_request_message}"
            )
            if self.signed_time_sliced_survey_request_message is not None
            else None
        )
        (
            out.append(
                f"signed_time_sliced_survey_response_message={self.signed_time_sliced_survey_response_message}"
            )
            if self.signed_time_sliced_survey_response_message is not None
            else None
        )
        (
            out.append(
                f"signed_time_sliced_survey_start_collecting_message={self.signed_time_sliced_survey_start_collecting_message}"
            )
            if self.signed_time_sliced_survey_start_collecting_message is not None
            else None
        )
        (
            out.append(
                f"signed_time_sliced_survey_stop_collecting_message={self.signed_time_sliced_survey_stop_collecting_message}"
            )
            if self.signed_time_sliced_survey_stop_collecting_message is not None
            else None
        )
        (
            out.append(f"q_set_hash={self.q_set_hash}")
            if self.q_set_hash is not None
            else None
        )
        out.append(f"q_set={self.q_set}") if self.q_set is not None else None
        out.append(f"envelope={self.envelope}") if self.envelope is not None else None
        (
            out.append(f"get_scp_ledger_seq={self.get_scp_ledger_seq}")
            if self.get_scp_ledger_seq is not None
            else None
        )
        (
            out.append(f"send_more_message={self.send_more_message}")
            if self.send_more_message is not None
            else None
        )
        (
            out.append(f"send_more_extended_message={self.send_more_extended_message}")
            if self.send_more_extended_message is not None
            else None
        )
        (
            out.append(f"flood_advert={self.flood_advert}")
            if self.flood_advert is not None
            else None
        )
        (
            out.append(f"flood_demand={self.flood_demand}")
            if self.flood_demand is not None
            else None
        )
        return f"<StellarMessage [{', '.join(out)}]>"



from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseFeeStatsCallBuilder"]


class BaseFeeStatsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`FeeStatsCallBuilder` pointed to server defined by horizon_url.

    See `Fee Stats <https://developers.stellar.org/api/aggregations/fee-stats/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "fee_stats"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .alpha_num4 import AlphaNum4
from .alpha_num12 import AlphaNum12
from .asset_type import AssetType

__all__ = ["Asset"]


class Asset:
    """
    XDR Source Code::

        union Asset switch (AssetType type)
        {
        case ASSET_TYPE_NATIVE: // Not credit
            void;

        case ASSET_TYPE_CREDIT_ALPHANUM4:
            AlphaNum4 alphaNum4;

        case ASSET_TYPE_CREDIT_ALPHANUM12:
            AlphaNum12 alphaNum12;

            // add other asset types here in the future
        };
    """

    def __init__(
        self,
        type: AssetType,
        alpha_num4: AlphaNum4 = None,
        alpha_num12: AlphaNum12 = None,
    ) -> None:
        self.type = type
        self.alpha_num4 = alpha_num4
        self.alpha_num12 = alpha_num12

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == AssetType.ASSET_TYPE_NATIVE:
            return
        if self.type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM4:
            if self.alpha_num4 is None:
                raise ValueError("alpha_num4 should not be None.")
            self.alpha_num4.pack(packer)
            return
        if self.type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM12:
            if self.alpha_num12 is None:
                raise ValueError("alpha_num12 should not be None.")
            self.alpha_num12.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Asset:
        type = AssetType.unpack(unpacker)
        if type == AssetType.ASSET_TYPE_NATIVE:
            return cls(type=type)
        if type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM4:
            alpha_num4 = AlphaNum4.unpack(unpacker)
            return cls(type=type, alpha_num4=alpha_num4)
        if type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM12:
            alpha_num12 = AlphaNum12.unpack(unpacker)
            return cls(type=type, alpha_num12=alpha_num12)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Asset:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Asset:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.alpha_num4,
                self.alpha_num12,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.alpha_num4 == other.alpha_num4
            and self.alpha_num12 == other.alpha_num12
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"alpha_num4={self.alpha_num4}")
            if self.alpha_num4 is not None
            else None
        )
        (
            out.append(f"alpha_num12={self.alpha_num12}")
            if self.alpha_num12 is not None
            else None
        )
        return f"<Asset [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .peer_stats import PeerStats
from .uint32 import Uint32

__all__ = ["TimeSlicedPeerData"]


class TimeSlicedPeerData:
    """
    XDR Source Code::

        struct TimeSlicedPeerData
        {
            PeerStats peerStats;
            uint32 averageLatencyMs;
        };
    """

    def __init__(
        self,
        peer_stats: PeerStats,
        average_latency_ms: Uint32,
    ) -> None:
        self.peer_stats = peer_stats
        self.average_latency_ms = average_latency_ms

    def pack(self, packer: Packer) -> None:
        self.peer_stats.pack(packer)
        self.average_latency_ms.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TimeSlicedPeerData:
        peer_stats = PeerStats.unpack(unpacker)
        average_latency_ms = Uint32.unpack(unpacker)
        return cls(
            peer_stats=peer_stats,
            average_latency_ms=average_latency_ms,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TimeSlicedPeerData:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TimeSlicedPeerData:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.peer_stats,
                self.average_latency_ms,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.peer_stats == other.peer_stats
            and self.average_latency_ms == other.average_latency_ms
        )

    def __repr__(self):
        out = [
            f"peer_stats={self.peer_stats}",
            f"average_latency_ms={self.average_latency_ms}",
        ]
        return f"<TimeSlicedPeerData [{', '.join(out)}]>"



from . import scval
from .__version__ import (
    __author__,
    __author_email__,
    __description__,
    __issues__,
    __license__,
    __title__,
    __url__,
    __version__,
)
from .account import *
from .address import *
from .asset import *
from .client.requests_client import RequestsClient
from .decorated_signature import *
from .fee_bump_transaction import *
from .fee_bump_transaction_envelope import *
from .helpers import *
from .keypair import *
from .ledger_bounds import *
from .liquidity_pool_asset import *
from .liquidity_pool_id import *
from .memo import *
from .muxed_account import *
from .network import *
from .operation import *
from .preconditions import *
from .price import *
from .server import *
from .signer import *
from .signer_key import *
from .soroban_data_builder import *
from .soroban_server import *
from .strkey import *
from .time_bounds import *
from .transaction import *
from .transaction_builder import *
from .transaction_envelope import *

# aiohttp required
try:
    from .client.aiohttp_client import AiohttpClient
    from .server_async import ServerAsync
    from .soroban_server_async import SorobanServerAsync
except ImportError:
    pass



from ...call_builder.base import BaseFeeStatsCallBuilder
from ...call_builder.call_builder_async.base_call_builder import BaseCallBuilder
from ...client.base_async_client import BaseAsyncClient

__all__ = ["FeeStatsCallBuilder"]


class FeeStatsCallBuilder(BaseCallBuilder, BaseFeeStatsCallBuilder):
    """Creates a new :class:`FeeStatsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.ServerAsync.fee_stats`.

    See `Fee Stats <https://developers.stellar.org/api/aggregations/fee-stats/>`__ for more information.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(self, horizon_url: str, client: BaseAsyncClient) -> None:
        super().__init__(horizon_url=horizon_url, client=client)

    async def stream(
        self,
    ):
        """This endpoint does not support streaming."""
        raise NotImplementedError("Streaming is not supported.")



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .base import String
from .constants import *
from .sc_spec_function_input_v0 import SCSpecFunctionInputV0
from .sc_spec_type_def import SCSpecTypeDef
from .sc_symbol import SCSymbol

__all__ = ["SCSpecFunctionV0"]


class SCSpecFunctionV0:
    """
    XDR Source Code::

        struct SCSpecFunctionV0
        {
            string doc<SC_SPEC_DOC_LIMIT>;
            SCSymbol name;
            SCSpecFunctionInputV0 inputs<10>;
            SCSpecTypeDef outputs<1>;
        };
    """

    def __init__(
        self,
        doc: bytes,
        name: SCSymbol,
        inputs: List[SCSpecFunctionInputV0],
        outputs: List[SCSpecTypeDef],
    ) -> None:
        _expect_max_length = 10
        if inputs and len(inputs) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `inputs` should be {_expect_max_length}, but got {len(inputs)}."
            )
        _expect_max_length = 1
        if outputs and len(outputs) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `outputs` should be {_expect_max_length}, but got {len(outputs)}."
            )
        self.doc = doc
        self.name = name
        self.inputs = inputs
        self.outputs = outputs

    def pack(self, packer: Packer) -> None:
        String(self.doc, SC_SPEC_DOC_LIMIT).pack(packer)
        self.name.pack(packer)
        packer.pack_uint(len(self.inputs))
        for inputs_item in self.inputs:
            inputs_item.pack(packer)
        packer.pack_uint(len(self.outputs))
        for outputs_item in self.outputs:
            outputs_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecFunctionV0:
        doc = String.unpack(unpacker)
        name = SCSymbol.unpack(unpacker)
        length = unpacker.unpack_uint()
        inputs = []
        for _ in range(length):
            inputs.append(SCSpecFunctionInputV0.unpack(unpacker))
        length = unpacker.unpack_uint()
        outputs = []
        for _ in range(length):
            outputs.append(SCSpecTypeDef.unpack(unpacker))
        return cls(
            doc=doc,
            name=name,
            inputs=inputs,
            outputs=outputs,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecFunctionV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecFunctionV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.doc,
                self.name,
                self.inputs,
                self.outputs,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.doc == other.doc
            and self.name == other.name
            and self.inputs == other.inputs
            and self.outputs == other.outputs
        )

    def __repr__(self):
        out = [
            f"doc={self.doc}",
            f"name={self.name}",
            f"inputs={self.inputs}",
            f"outputs={self.outputs}",
        ]
        return f"<SCSpecFunctionV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .contract_executable import ContractExecutable
from .contract_id_preimage import ContractIDPreimage
from .sc_val import SCVal

__all__ = ["CreateContractArgsV2"]


class CreateContractArgsV2:
    """
    XDR Source Code::

        struct CreateContractArgsV2
        {
            ContractIDPreimage contractIDPreimage;
            ContractExecutable executable;
            // Arguments of the contract's constructor.
            SCVal constructorArgs<>;
        };
    """

    def __init__(
        self,
        contract_id_preimage: ContractIDPreimage,
        executable: ContractExecutable,
        constructor_args: List[SCVal],
    ) -> None:
        _expect_max_length = 4294967295
        if constructor_args and len(constructor_args) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `constructor_args` should be {_expect_max_length}, but got {len(constructor_args)}."
            )
        self.contract_id_preimage = contract_id_preimage
        self.executable = executable
        self.constructor_args = constructor_args

    def pack(self, packer: Packer) -> None:
        self.contract_id_preimage.pack(packer)
        self.executable.pack(packer)
        packer.pack_uint(len(self.constructor_args))
        for constructor_args_item in self.constructor_args:
            constructor_args_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> CreateContractArgsV2:
        contract_id_preimage = ContractIDPreimage.unpack(unpacker)
        executable = ContractExecutable.unpack(unpacker)
        length = unpacker.unpack_uint()
        constructor_args = []
        for _ in range(length):
            constructor_args.append(SCVal.unpack(unpacker))
        return cls(
            contract_id_preimage=contract_id_preimage,
            executable=executable,
            constructor_args=constructor_args,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> CreateContractArgsV2:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> CreateContractArgsV2:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.contract_id_preimage,
                self.executable,
                self.constructor_args,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.contract_id_preimage == other.contract_id_preimage
            and self.executable == other.executable
            and self.constructor_args == other.constructor_args
        )

    def __repr__(self):
        out = [
            f"contract_id_preimage={self.contract_id_preimage}",
            f"executable={self.executable}",
            f"constructor_args={self.constructor_args}",
        ]
        return f"<CreateContractArgsV2 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["ContractCostType"]


class ContractCostType(IntEnum):
    """
    XDR Source Code::

        enum ContractCostType {
            // Cost of running 1 wasm instruction
            WasmInsnExec = 0,
            // Cost of allocating a slice of memory (in bytes)
            MemAlloc = 1,
            // Cost of copying a slice of bytes into a pre-allocated memory
            MemCpy = 2,
            // Cost of comparing two slices of memory
            MemCmp = 3,
            // Cost of a host function dispatch, not including the actual work done by
            // the function nor the cost of VM invocation machinary
            DispatchHostFunction = 4,
            // Cost of visiting a host object from the host object storage. Exists to
            // make sure some baseline cost coverage, i.e. repeatly visiting objects
            // by the guest will always incur some charges.
            VisitObject = 5,
            // Cost of serializing an xdr object to bytes
            ValSer = 6,
            // Cost of deserializing an xdr object from bytes
            ValDeser = 7,
            // Cost of computing the sha256 hash from bytes
            ComputeSha256Hash = 8,
            // Cost of computing the ed25519 pubkey from bytes
            ComputeEd25519PubKey = 9,
            // Cost of verifying ed25519 signature of a payload.
            VerifyEd25519Sig = 10,
            // Cost of instantiation a VM from wasm bytes code.
            VmInstantiation = 11,
            // Cost of instantiation a VM from a cached state.
            VmCachedInstantiation = 12,
            // Cost of invoking a function on the VM. If the function is a host function,
            // additional cost will be covered by `DispatchHostFunction`.
            InvokeVmFunction = 13,
            // Cost of computing a keccak256 hash from bytes.
            ComputeKeccak256Hash = 14,
            // Cost of decoding an ECDSA signature computed from a 256-bit prime modulus
            // curve (e.g. secp256k1 and secp256r1)
            DecodeEcdsaCurve256Sig = 15,
            // Cost of recovering an ECDSA secp256k1 key from a signature.
            RecoverEcdsaSecp256k1Key = 16,
            // Cost of int256 addition (`+`) and subtraction (`-`) operations
            Int256AddSub = 17,
            // Cost of int256 multiplication (`*`) operation
            Int256Mul = 18,
            // Cost of int256 division (`/`) operation
            Int256Div = 19,
            // Cost of int256 power (`exp`) operation
            Int256Pow = 20,
            // Cost of int256 shift (`shl`, `shr`) operation
            Int256Shift = 21,
            // Cost of drawing random bytes using a ChaCha20 PRNG
            ChaCha20DrawBytes = 22,

            // Cost of parsing wasm bytes that only encode instructions.
            ParseWasmInstructions = 23,
            // Cost of parsing a known number of wasm functions.
            ParseWasmFunctions = 24,
            // Cost of parsing a known number of wasm globals.
            ParseWasmGlobals = 25,
            // Cost of parsing a known number of wasm table entries.
            ParseWasmTableEntries = 26,
            // Cost of parsing a known number of wasm types.
            ParseWasmTypes = 27,
            // Cost of parsing a known number of wasm data segments.
            ParseWasmDataSegments = 28,
            // Cost of parsing a known number of wasm element segments.
            ParseWasmElemSegments = 29,
            // Cost of parsing a known number of wasm imports.
            ParseWasmImports = 30,
            // Cost of parsing a known number of wasm exports.
            ParseWasmExports = 31,
            // Cost of parsing a known number of data segment bytes.
            ParseWasmDataSegmentBytes = 32,

            // Cost of instantiating wasm bytes that only encode instructions.
            InstantiateWasmInstructions = 33,
            // Cost of instantiating a known number of wasm functions.
            InstantiateWasmFunctions = 34,
            // Cost of instantiating a known number of wasm globals.
            InstantiateWasmGlobals = 35,
            // Cost of instantiating a known number of wasm table entries.
            InstantiateWasmTableEntries = 36,
            // Cost of instantiating a known number of wasm types.
            InstantiateWasmTypes = 37,
            // Cost of instantiating a known number of wasm data segments.
            InstantiateWasmDataSegments = 38,
            // Cost of instantiating a known number of wasm element segments.
            InstantiateWasmElemSegments = 39,
            // Cost of instantiating a known number of wasm imports.
            InstantiateWasmImports = 40,
            // Cost of instantiating a known number of wasm exports.
            InstantiateWasmExports = 41,
            // Cost of instantiating a known number of data segment bytes.
            InstantiateWasmDataSegmentBytes = 42,

            // Cost of decoding a bytes array representing an uncompressed SEC-1 encoded
            // point on a 256-bit elliptic curve
            Sec1DecodePointUncompressed = 43,
            // Cost of verifying an ECDSA Secp256r1 signature
            VerifyEcdsaSecp256r1Sig = 44,

            // Cost of encoding a BLS12-381 Fp (base field element)
            Bls12381EncodeFp = 45,
            // Cost of decoding a BLS12-381 Fp (base field element)
            Bls12381DecodeFp = 46,
            // Cost of checking a G1 point lies on the curve
            Bls12381G1CheckPointOnCurve = 47,
            // Cost of checking a G1 point belongs to the correct subgroup
            Bls12381G1CheckPointInSubgroup = 48,
            // Cost of checking a G2 point lies on the curve
            Bls12381G2CheckPointOnCurve = 49,
            // Cost of checking a G2 point belongs to the correct subgroup
            Bls12381G2CheckPointInSubgroup = 50,
            // Cost of converting a BLS12-381 G1 point from projective to affine coordinates
            Bls12381G1ProjectiveToAffine = 51,
            // Cost of converting a BLS12-381 G2 point from projective to affine coordinates
            Bls12381G2ProjectiveToAffine = 52,
            // Cost of performing BLS12-381 G1 point addition
            Bls12381G1Add = 53,
            // Cost of performing BLS12-381 G1 scalar multiplication
            Bls12381G1Mul = 54,
            // Cost of performing BLS12-381 G1 multi-scalar multiplication (MSM)
            Bls12381G1Msm = 55,
            // Cost of mapping a BLS12-381 Fp field element to a G1 point
            Bls12381MapFpToG1 = 56,
            // Cost of hashing to a BLS12-381 G1 point
            Bls12381HashToG1 = 57,
            // Cost of performing BLS12-381 G2 point addition
            Bls12381G2Add = 58,
            // Cost of performing BLS12-381 G2 scalar multiplication
            Bls12381G2Mul = 59,
            // Cost of performing BLS12-381 G2 multi-scalar multiplication (MSM)
            Bls12381G2Msm = 60,
            // Cost of mapping a BLS12-381 Fp2 field element to a G2 point
            Bls12381MapFp2ToG2 = 61,
            // Cost of hashing to a BLS12-381 G2 point
            Bls12381HashToG2 = 62,
            // Cost of performing BLS12-381 pairing operation
            Bls12381Pairing = 63,
            // Cost of converting a BLS12-381 scalar element from U256
            Bls12381FrFromU256 = 64,
            // Cost of converting a BLS12-381 scalar element to U256
            Bls12381FrToU256 = 65,
            // Cost of performing BLS12-381 scalar element addition/subtraction
            Bls12381FrAddSub = 66,
            // Cost of performing BLS12-381 scalar element multiplication
            Bls12381FrMul = 67,
            // Cost of performing BLS12-381 scalar element exponentiation
            Bls12381FrPow = 68,
            // Cost of performing BLS12-381 scalar element inversion
            Bls12381FrInv = 69
        };
    """

    WasmInsnExec = 0
    MemAlloc = 1
    MemCpy = 2
    MemCmp = 3
    DispatchHostFunction = 4
    VisitObject = 5
    ValSer = 6
    ValDeser = 7
    ComputeSha256Hash = 8
    ComputeEd25519PubKey = 9
    VerifyEd25519Sig = 10
    VmInstantiation = 11
    VmCachedInstantiation = 12
    InvokeVmFunction = 13
    ComputeKeccak256Hash = 14
    DecodeEcdsaCurve256Sig = 15
    RecoverEcdsaSecp256k1Key = 16
    Int256AddSub = 17
    Int256Mul = 18
    Int256Div = 19
    Int256Pow = 20
    Int256Shift = 21
    ChaCha20DrawBytes = 22
    ParseWasmInstructions = 23
    ParseWasmFunctions = 24
    ParseWasmGlobals = 25
    ParseWasmTableEntries = 26
    ParseWasmTypes = 27
    ParseWasmDataSegments = 28
    ParseWasmElemSegments = 29
    ParseWasmImports = 30
    ParseWasmExports = 31
    ParseWasmDataSegmentBytes = 32
    InstantiateWasmInstructions = 33
    InstantiateWasmFunctions = 34
    InstantiateWasmGlobals = 35
    InstantiateWasmTableEntries = 36
    InstantiateWasmTypes = 37
    InstantiateWasmDataSegments = 38
    InstantiateWasmElemSegments = 39
    InstantiateWasmImports = 40
    InstantiateWasmExports = 41
    InstantiateWasmDataSegmentBytes = 42
    Sec1DecodePointUncompressed = 43
    VerifyEcdsaSecp256r1Sig = 44
    Bls12381EncodeFp = 45
    Bls12381DecodeFp = 46
    Bls12381G1CheckPointOnCurve = 47
    Bls12381G1CheckPointInSubgroup = 48
    Bls12381G2CheckPointOnCurve = 49
    Bls12381G2CheckPointInSubgroup = 50
    Bls12381G1ProjectiveToAffine = 51
    Bls12381G2ProjectiveToAffine = 52
    Bls12381G1Add = 53
    Bls12381G1Mul = 54
    Bls12381G1Msm = 55
    Bls12381MapFpToG1 = 56
    Bls12381HashToG1 = 57
    Bls12381G2Add = 58
    Bls12381G2Mul = 59
    Bls12381G2Msm = 60
    Bls12381MapFp2ToG2 = 61
    Bls12381HashToG2 = 62
    Bls12381Pairing = 63
    Bls12381FrFromU256 = 64
    Bls12381FrToU256 = 65
    Bls12381FrAddSub = 66
    Bls12381FrMul = 67
    Bls12381FrPow = 68
    Bls12381FrInv = 69

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractCostType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractCostType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractCostType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



import warnings
from enum import IntFlag
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..keypair import Keypair
from ..muxed_account import MuxedAccount
from ..signer import Signer
from ..strkey import StrKey
from ..utils import raise_if_not_valid_ed25519_public_key
from .operation import Operation

__all__ = ["AuthorizationFlag", "SetOptions"]


class AuthorizationFlag(IntFlag):
    """Indicates which flags to set. For details about the flags,
    please refer to the `Control Access to an Asset - Flag <https://developers.stellar.org/docs/issuing-assets/control-asset-access/#flags>`__.
    """

    AUTHORIZATION_REQUIRED = 1
    AUTHORIZATION_REVOCABLE = 2
    AUTHORIZATION_IMMUTABLE = 4
    AUTHORIZATION_CLAWBACK_ENABLED = 8


class SetOptions(Operation):
    """The :class:`SetOptions` object, which represents a SetOptions operation
    on Stellar's network.

    This operation sets the options for an account.

    For more information on the signing options, please refer to the `multi-sig
    doc <https://developers.stellar.org/docs/glossary/multisig/>`_.

    When updating signers or other thresholds, the threshold of this operation
    is high.

    Threshold: Medium or High

    See `Set Options <https://developers.stellar.org/docs/start/list-of-operations/#set-options>`_ for more information.

    :param inflation_dest: Account of the inflation destination.
    :param clear_flags: Indicates which flags to clear. For details about the flags,
        please refer to the `Control Access to an Asset - Flag <https://developers.stellar.org/docs/issuing-assets/control-asset-access/#flags>`__.
        The `bit mask <https://en.wikipedia.org/wiki/Bit_field>`_ integer subtracts from the existing flags of the account.
        This allows for setting specific bits without knowledge of existing flags, you can also use
        :class:`stellar_sdk.operation.set_options.AuthorizationFlag`

        * AUTHORIZATION_REQUIRED = 1
        * AUTHORIZATION_REVOCABLE = 2
        * AUTHORIZATION_IMMUTABLE = 4
        * AUTHORIZATION_CLAWBACK_ENABLED = 8

    :param set_flags: Indicates which flags to set. For details about the flags,
        please refer to the `Control Access to an Asset - Flag <https://developers.stellar.org/docs/issuing-assets/control-asset-access/#flags>`__.
        The bit mask integer adds onto the existing flags of the account.
        This allows for setting specific bits without knowledge of existing flags, you can also use
        :class:`stellar_sdk.operation.set_options.AuthorizationFlag`

        * AUTHORIZATION_REQUIRED = 1
        * AUTHORIZATION_REVOCABLE = 2
        * AUTHORIZATION_IMMUTABLE = 4
        * AUTHORIZATION_CLAWBACK_ENABLED = 8

    :param master_weight: A number from 0-255 (inclusive) representing the weight of the master key.
        If the weight of the master key is updated to 0, it is effectively disabled.
    :param low_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
        operations it performs that have `a low threshold <https://developers.stellar.org/docs/glossary/multisig/>`_.
    :param med_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
        operations it performs that have `a medium threshold <https://developers.stellar.org/docs/glossary/multisig/>`_.
    :param high_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
        operations it performs that have `a high threshold <https://developers.stellar.org/docs/glossary/multisig/>`_.
    :param home_domain: sets the home domain used for
        reverse `federation <https://developers.stellar.org/docs/glossary/federation/>`_ lookup.
    :param signer: Add, update, or remove a signer from the account.
    :param source: The source account for the operation. Defaults to the transaction's source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.SET_OPTIONS
    )

    def __init__(
        self,
        inflation_dest: str = None,
        clear_flags: Union[int, AuthorizationFlag] = None,
        set_flags: Union[int, AuthorizationFlag] = None,
        master_weight: int = None,
        low_threshold: int = None,
        med_threshold: int = None,
        high_threshold: int = None,
        signer: Signer = None,
        home_domain: str = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        if set_flags is not None and not isinstance(set_flags, AuthorizationFlag):
            warnings.warn(
                "`set_flags` is a int, we recommend using AuthorizationFlag.",
                DeprecationWarning,
            )
            set_flags = AuthorizationFlag(set_flags)

        if clear_flags is not None and not isinstance(clear_flags, AuthorizationFlag):
            warnings.warn(
                "`clear_flags` is a int, we recommend using AuthorizationFlag.",
                DeprecationWarning,
            )
            clear_flags = AuthorizationFlag(clear_flags)

        self.inflation_dest = inflation_dest
        self.clear_flags: Optional[AuthorizationFlag] = clear_flags
        self.set_flags: Optional[AuthorizationFlag] = set_flags
        self.master_weight: Optional[int] = master_weight
        self.low_threshold: Optional[int] = low_threshold
        self.med_threshold: Optional[int] = med_threshold
        self.high_threshold: Optional[int] = high_threshold
        self.home_domain: Optional[str] = home_domain
        self.signer: Optional[Signer] = signer
        if self.inflation_dest is not None:
            raise_if_not_valid_ed25519_public_key(self.inflation_dest, "inflation_dest")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        inflation_dest = (
            Keypair.from_public_key(self.inflation_dest).xdr_account_id()
            if self.inflation_dest is not None
            else None
        )
        home_domain = (
            stellar_xdr.String32(bytes(self.home_domain, encoding="utf-8"))
            if self.home_domain is not None
            else None
        )
        clear_flags = (
            None
            if self.clear_flags is None
            else stellar_xdr.Uint32(self.clear_flags.value)
        )
        set_flags = (
            None if self.set_flags is None else stellar_xdr.Uint32(self.set_flags.value)
        )
        master_weight = (
            None
            if self.master_weight is None
            else stellar_xdr.Uint32(self.master_weight)
        )
        low_threshold = (
            None
            if self.low_threshold is None
            else stellar_xdr.Uint32(self.low_threshold)
        )
        med_threshold = (
            None
            if self.med_threshold is None
            else stellar_xdr.Uint32(self.med_threshold)
        )
        high_threshold = (
            None
            if self.high_threshold is None
            else stellar_xdr.Uint32(self.high_threshold)
        )
        signer = None if self.signer is None else self.signer.to_xdr_object()

        set_options_op = stellar_xdr.SetOptionsOp(
            inflation_dest,
            clear_flags,
            set_flags,
            master_weight,
            low_threshold,
            med_threshold,
            high_threshold,
            home_domain,
            signer,
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, set_options_op=set_options_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object) -> "SetOptions":
        """Creates a :class:`SetOptions` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)

        inflation_dest = None
        if xdr_object.body.set_options_op.inflation_dest:
            inflation_dest = StrKey.encode_ed25519_public_key(
                xdr_object.body.set_options_op.inflation_dest.account_id.ed25519.uint256
            )

        clear_flags_xdr = xdr_object.body.set_options_op.clear_flags
        set_flags_xdr = xdr_object.body.set_options_op.set_flags
        master_weight_xdr = xdr_object.body.set_options_op.master_weight
        low_threshold_xdr = xdr_object.body.set_options_op.low_threshold
        med_threshold_xdr = xdr_object.body.set_options_op.med_threshold
        high_threshold_xdr = xdr_object.body.set_options_op.high_threshold
        home_domain_xdr = xdr_object.body.set_options_op.home_domain
        signer_xdr_object = xdr_object.body.set_options_op.signer

        clear_flags = None if clear_flags_xdr is None else clear_flags_xdr.uint32
        set_flags = None if set_flags_xdr is None else set_flags_xdr.uint32
        master_weight = None if master_weight_xdr is None else master_weight_xdr.uint32
        low_threshold = None if low_threshold_xdr is None else low_threshold_xdr.uint32
        med_threshold = None if med_threshold_xdr is None else med_threshold_xdr.uint32
        high_threshold = (
            None if high_threshold_xdr is None else high_threshold_xdr.uint32
        )
        home_domain = None if home_domain_xdr is None else home_domain_xdr.string32
        signer = (
            None
            if signer_xdr_object is None
            else Signer.from_xdr_object(signer_xdr_object)
        )

        if home_domain is not None:
            home_domain = home_domain.decode("utf-8")

        op = cls(
            inflation_dest=inflation_dest,
            clear_flags=clear_flags,
            set_flags=set_flags,
            master_weight=master_weight,
            low_threshold=low_threshold,
            med_threshold=med_threshold,
            high_threshold=high_threshold,
            home_domain=home_domain,
            signer=signer,
            source=source,
        )
        return op

    def __repr__(self):
        return (
            f"<SetOptions [inflation_dest={self.inflation_dest}, "
            f"clear_flags={self.clear_flags}, "
            f"set_flags={self.set_flags}, "
            f"master_weight={self.master_weight}, "
            f"low_threshold={self.low_threshold}, "
            f"med_threshold={self.med_threshold}, "
            f"high_threshold={self.high_threshold}, "
            f"signer={self.signer}, "
            f"home_domain={self.home_domain}, "
            f"source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .host_function import HostFunction
from .soroban_authorization_entry import SorobanAuthorizationEntry

__all__ = ["InvokeHostFunctionOp"]


class InvokeHostFunctionOp:
    """
    XDR Source Code::

        struct InvokeHostFunctionOp
        {
            // Host function to invoke.
            HostFunction hostFunction;
            // Per-address authorizations for this host function.
            SorobanAuthorizationEntry auth<>;
        };
    """

    def __init__(
        self,
        host_function: HostFunction,
        auth: List[SorobanAuthorizationEntry],
    ) -> None:
        _expect_max_length = 4294967295
        if auth and len(auth) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `auth` should be {_expect_max_length}, but got {len(auth)}."
            )
        self.host_function = host_function
        self.auth = auth

    def pack(self, packer: Packer) -> None:
        self.host_function.pack(packer)
        packer.pack_uint(len(self.auth))
        for auth_item in self.auth:
            auth_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> InvokeHostFunctionOp:
        host_function = HostFunction.unpack(unpacker)
        length = unpacker.unpack_uint()
        auth = []
        for _ in range(length):
            auth.append(SorobanAuthorizationEntry.unpack(unpacker))
        return cls(
            host_function=host_function,
            auth=auth,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> InvokeHostFunctionOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> InvokeHostFunctionOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.host_function,
                self.auth,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.host_function == other.host_function and self.auth == other.auth

    def __repr__(self):
        out = [
            f"host_function={self.host_function}",
            f"auth={self.auth}",
        ]
        return f"<InvokeHostFunctionOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["SCSpecType"]


class SCSpecType(IntEnum):
    """
    XDR Source Code::

        enum SCSpecType
        {
            SC_SPEC_TYPE_VAL = 0,

            // Types with no parameters.
            SC_SPEC_TYPE_BOOL = 1,
            SC_SPEC_TYPE_VOID = 2,
            SC_SPEC_TYPE_ERROR = 3,
            SC_SPEC_TYPE_U32 = 4,
            SC_SPEC_TYPE_I32 = 5,
            SC_SPEC_TYPE_U64 = 6,
            SC_SPEC_TYPE_I64 = 7,
            SC_SPEC_TYPE_TIMEPOINT = 8,
            SC_SPEC_TYPE_DURATION = 9,
            SC_SPEC_TYPE_U128 = 10,
            SC_SPEC_TYPE_I128 = 11,
            SC_SPEC_TYPE_U256 = 12,
            SC_SPEC_TYPE_I256 = 13,
            SC_SPEC_TYPE_BYTES = 14,
            SC_SPEC_TYPE_STRING = 16,
            SC_SPEC_TYPE_SYMBOL = 17,
            SC_SPEC_TYPE_ADDRESS = 19,

            // Types with parameters.
            SC_SPEC_TYPE_OPTION = 1000,
            SC_SPEC_TYPE_RESULT = 1001,
            SC_SPEC_TYPE_VEC = 1002,
            SC_SPEC_TYPE_MAP = 1004,
            SC_SPEC_TYPE_TUPLE = 1005,
            SC_SPEC_TYPE_BYTES_N = 1006,

            // User defined types.
            SC_SPEC_TYPE_UDT = 2000
        };
    """

    SC_SPEC_TYPE_VAL = 0
    SC_SPEC_TYPE_BOOL = 1
    SC_SPEC_TYPE_VOID = 2
    SC_SPEC_TYPE_ERROR = 3
    SC_SPEC_TYPE_U32 = 4
    SC_SPEC_TYPE_I32 = 5
    SC_SPEC_TYPE_U64 = 6
    SC_SPEC_TYPE_I64 = 7
    SC_SPEC_TYPE_TIMEPOINT = 8
    SC_SPEC_TYPE_DURATION = 9
    SC_SPEC_TYPE_U128 = 10
    SC_SPEC_TYPE_I128 = 11
    SC_SPEC_TYPE_U256 = 12
    SC_SPEC_TYPE_I256 = 13
    SC_SPEC_TYPE_BYTES = 14
    SC_SPEC_TYPE_STRING = 16
    SC_SPEC_TYPE_SYMBOL = 17
    SC_SPEC_TYPE_ADDRESS = 19
    SC_SPEC_TYPE_OPTION = 1000
    SC_SPEC_TYPE_RESULT = 1001
    SC_SPEC_TYPE_VEC = 1002
    SC_SPEC_TYPE_MAP = 1004
    SC_SPEC_TYPE_TUPLE = 1005
    SC_SPEC_TYPE_BYTES_N = 1006
    SC_SPEC_TYPE_UDT = 2000

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .int64 import Int64
from .price import Price

__all__ = ["ManageSellOfferOp"]


class ManageSellOfferOp:
    """
    XDR Source Code::

        struct ManageSellOfferOp
        {
            Asset selling;
            Asset buying;
            int64 amount; // amount being sold. if set to 0, delete the offer
            Price price;  // price of thing being sold in terms of what you are buying

            // 0=create a new offer, otherwise edit an existing offer
            int64 offerID;
        };
    """

    def __init__(
        self,
        selling: Asset,
        buying: Asset,
        amount: Int64,
        price: Price,
        offer_id: Int64,
    ) -> None:
        self.selling = selling
        self.buying = buying
        self.amount = amount
        self.price = price
        self.offer_id = offer_id

    def pack(self, packer: Packer) -> None:
        self.selling.pack(packer)
        self.buying.pack(packer)
        self.amount.pack(packer)
        self.price.pack(packer)
        self.offer_id.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ManageSellOfferOp:
        selling = Asset.unpack(unpacker)
        buying = Asset.unpack(unpacker)
        amount = Int64.unpack(unpacker)
        price = Price.unpack(unpacker)
        offer_id = Int64.unpack(unpacker)
        return cls(
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ManageSellOfferOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ManageSellOfferOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.selling,
                self.buying,
                self.amount,
                self.price,
                self.offer_id,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.selling == other.selling
            and self.buying == other.buying
            and self.amount == other.amount
            and self.price == other.price
            and self.offer_id == other.offer_id
        )

    def __repr__(self):
        out = [
            f"selling={self.selling}",
            f"buying={self.buying}",
            f"amount={self.amount}",
            f"price={self.price}",
            f"offer_id={self.offer_id}",
        ]
        return f"<ManageSellOfferOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .manage_offer_effect import ManageOfferEffect
from .offer_entry import OfferEntry

__all__ = ["ManageOfferSuccessResultOffer"]


class ManageOfferSuccessResultOffer:
    """
    XDR Source Code::

        union switch (ManageOfferEffect effect)
            {
            case MANAGE_OFFER_CREATED:
            case MANAGE_OFFER_UPDATED:
                OfferEntry offer;
            case MANAGE_OFFER_DELETED:
                void;
            }
    """

    def __init__(
        self,
        effect: ManageOfferEffect,
        offer: OfferEntry = None,
    ) -> None:
        self.effect = effect
        self.offer = offer

    def pack(self, packer: Packer) -> None:
        self.effect.pack(packer)
        if self.effect == ManageOfferEffect.MANAGE_OFFER_CREATED:
            if self.offer is None:
                raise ValueError("offer should not be None.")
            self.offer.pack(packer)
            return
        if self.effect == ManageOfferEffect.MANAGE_OFFER_UPDATED:
            if self.offer is None:
                raise ValueError("offer should not be None.")
            self.offer.pack(packer)
            return
        if self.effect == ManageOfferEffect.MANAGE_OFFER_DELETED:
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ManageOfferSuccessResultOffer:
        effect = ManageOfferEffect.unpack(unpacker)
        if effect == ManageOfferEffect.MANAGE_OFFER_CREATED:
            offer = OfferEntry.unpack(unpacker)
            return cls(effect=effect, offer=offer)
        if effect == ManageOfferEffect.MANAGE_OFFER_UPDATED:
            offer = OfferEntry.unpack(unpacker)
            return cls(effect=effect, offer=offer)
        if effect == ManageOfferEffect.MANAGE_OFFER_DELETED:
            return cls(effect=effect)
        return cls(effect=effect)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ManageOfferSuccessResultOffer:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ManageOfferSuccessResultOffer:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.effect,
                self.offer,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.effect == other.effect and self.offer == other.offer

    def __repr__(self):
        out = []
        out.append(f"effect={self.effect}")
        out.append(f"offer={self.offer}") if self.offer is not None else None
        return f"<ManageOfferSuccessResultOffer [{', '.join(out)}]>"



from typing import Union

from ...asset import Asset
from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseTradesCallBuilder"]


class BaseTradesCallBuilder(BaseCallBuilder):
    """Creates a new :class:`TradesCallBuilder` pointed to server defined by horizon_url.

    See `List All Trades <https://developers.stellar.org/api/resources/trades/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "trades"

    def for_asset_pair(self, base: Asset, counter: Asset):
        """Filter trades for a specific asset pair (orderbook)

        See `List All Trades <https://developers.stellar.org/api/resources/trades/list/>`__ for more information.

        :param base: base asset
        :param counter: counter asset
        :return: current TradesCallBuilder instance
        """
        params = {
            "base_asset_type": base.type,
            "base_asset_code": None if base.is_native() else base.code,
            "base_asset_issuer": base.issuer,
            "counter_asset_type": counter.type,
            "counter_asset_code": None if counter.is_native() else counter.code,
            "counter_asset_issuer": counter.issuer,
        }
        self._add_query_params(params)
        return self

    def for_offer(self, offer_id: Union[int, str]):
        """Filter trades for a specific offer

        See `List All Trades <https://developers.stellar.org/api/resources/trades/list/>`__ for more information.

        :param offer_id: offer id
        :return: current TradesCallBuilder instance
        """
        self.endpoint = f"offers/{offer_id}/trades"
        return self

    def for_account(self, account_id: str):
        """Filter trades for a specific account

        See `Retrieve an Account's Trades <https://developers.stellar.org/api/resources/accounts/trades/>`__ for more information.

        :param account_id: account id
        :return: current TradesCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/trades"
        return self

    def for_trade_type(self, trade_type: str):
        """Filter trades for a specific trade type

        Horizon will reject requests which attempt to set
        `trade_type` to ``liquidity_pools`` when using the offer id filter.

        :param trade_type: trade type, the currently supported types are ``"orderbook"``, ``"liquidity_pool"`` and ``"all"``,
            defaults to ``"all"``.
        :return: current TradesCallBuilder instance
        """
        params = {"trade_type": trade_type}
        self._add_query_params(params)
        return self

    def for_liquidity_pool(self, liquidity_pool_id: str):
        """Filter trades for a specific liquidity pool.

        See `Liquidity Pools - Retrieve related Trades <https://developers.stellar.org/api/resources/liquiditypools/trades/>`__

        :param liquidity_pool_id: The ID of the liquidity pool in hex string.
        :return: current TradesCallBuilder instance
        """
        self.endpoint = f"liquidity_pools/{liquidity_pool_id}/trades"
        return self



import base64
import binascii
from enum import IntEnum
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..keypair import Keypair
from ..liquidity_pool_id import LiquidityPoolId
from ..muxed_account import MuxedAccount
from ..signer_key import SignerKey
from ..strkey import StrKey
from ..utils import (
    raise_if_not_valid_balance_id,
    raise_if_not_valid_ed25519_public_key,
    raise_if_not_valid_hash,
)
from .operation import Operation

__all__ = ["RevokeSponsorship"]


class RevokeSponsorshipType(IntEnum):
    """Currently supported RevokeSponsorship types."""

    ACCOUNT = 0
    TRUSTLINE = 1
    OFFER = 2
    DATA = 3
    CLAIMABLE_BALANCE = 4
    SIGNER = 5
    LIQUIDITY_POOL = 6


class TrustLine:
    def __init__(self, account_id: str, asset: Union[Asset, LiquidityPoolId]) -> None:
        self.account_id = account_id
        self.asset = asset
        raise_if_not_valid_ed25519_public_key(self.account_id, "account_id")

    def __hash__(self):
        return hash((self.account_id, self.asset))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.account_id == other.account_id and self.asset == other.asset

    def __repr__(self):
        return f"<TrustLine [account_id={self.account_id}, asset={self.asset}]>"


class Offer:
    def __init__(self, seller_id: str, offer_id: int) -> None:
        self.seller_id = seller_id
        self.offer_id = offer_id
        raise_if_not_valid_ed25519_public_key(self.seller_id, "seller_id")

    def __hash__(self):
        return hash((self.seller_id, self.offer_id))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.seller_id == other.seller_id and self.offer_id == other.offer_id

    def __repr__(self):
        return f"<Offer [seller_id={self.seller_id}, offer_id={self.offer_id}]>"


class Data:
    def __init__(self, account_id: str, data_name: str) -> None:
        self.account_id = account_id
        self.data_name = data_name
        raise_if_not_valid_ed25519_public_key(self.account_id, "account_id")

    def __hash__(self):
        return hash((self.account_id, self.data_name))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.account_id == other.account_id and self.data_name == other.data_name

    def __repr__(self):
        return f"<Data [account_id={self.account_id}, data_name={self.data_name}]>"


class Signer:
    def __init__(self, account_id: str, signer_key: SignerKey) -> None:
        self.account_id = account_id
        self.signer_key = signer_key
        raise_if_not_valid_ed25519_public_key(self.account_id, "account_id")

    def __hash__(self):
        return hash((self.account_id, self.signer_key))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.account_id == other.account_id and self.signer_key == other.signer_key
        )

    def __repr__(self):
        return f"<Signer [account_id={self.account_id}, signer_key={self.signer_key}]>"


class RevokeSponsorship(Operation):
    """The :class:`RevokeSponsorship` object, which represents a RevokeSponsorship
    operation on Stellar's network.

    The logic of this operation depends on the state of the source account.

    If the source account is not sponsored or is sponsored by the owner of the specified entry or sub-entry,
    then attempt to revoke the sponsorship. If the source account is sponsored, the next step depends on whether the
    entry is sponsored or not. If it is sponsored, attempt to transfer the sponsorship to the sponsor
    of the source account. If the entry is not sponsored, then establish the sponsorship.
    See `Sponsored Reserves <https://developers.stellar.org/docs/glossary/sponsored-reserves/>`_ for more information.

    Threshold: Medium

    See `Revoke Sponsorship <https://developers.stellar.org/docs/start/list-of-operations/#revoke-sponsorship>`_ for more information.

    :param revoke_sponsorship_type: The sponsored account id.
    :param account_id: The sponsored account ID.
    :param trustline: The sponsored trustline.
    :param offer: The sponsored offer.
    :param data: The sponsored data.
    :param claimable_balance_id: The sponsored claimable balance.
    :param signer: The sponsored signer.
    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.REVOKE_SPONSORSHIP
    )
    # TODO: Protocol 20?

    def __init__(
        self,
        revoke_sponsorship_type: RevokeSponsorshipType,
        account_id: Optional[str],
        trustline: Optional[TrustLine],
        offer: Optional[Offer],
        data: Optional[Data],
        claimable_balance_id: Optional[str],
        signer: Optional[Signer],
        liquidity_pool_id: Optional[str],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        self.revoke_sponsorship_type: RevokeSponsorshipType = revoke_sponsorship_type
        self.account_id: Optional[str] = account_id
        self.trustline: Optional[TrustLine] = trustline
        self.offer: Optional[Offer] = offer
        self.data: Optional[Data] = data
        self.claimable_balance_id: Optional[str] = claimable_balance_id
        self.signer: Optional[Signer] = signer
        self.liquidity_pool_id: Optional[str] = liquidity_pool_id

        if self.account_id is not None:
            raise_if_not_valid_ed25519_public_key(self.account_id, "account_id")
        if self.claimable_balance_id is not None:
            raise_if_not_valid_balance_id(
                self.claimable_balance_id, "claimable_balance_id"
            )
        if self.liquidity_pool_id is not None:
            raise_if_not_valid_hash(self.liquidity_pool_id, "liquidity_pool_id")

    @classmethod
    def revoke_account_sponsorship(
        cls, account_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ):
        """Create a **revoke sponsorship** operation for an account.

        :param account_id: The sponsored account ID.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for an account.
        """
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.ACCOUNT,
            account_id=account_id,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=None,
            liquidity_pool_id=None,
            source=source,
        )

    @classmethod
    def revoke_trustline_sponsorship(
        cls,
        account_id: str,
        asset: Union[Asset, LiquidityPoolId],
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        """Create a **revoke sponsorship** operation for a trustline.

        :param account_id: The account ID which owns the trustline.
        :param asset: The asset in the trustline.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for a trustline.
        """
        trustline = TrustLine(account_id=account_id, asset=asset)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.TRUSTLINE,
            account_id=None,
            trustline=trustline,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=None,
            liquidity_pool_id=None,
            source=source,
        )

    @classmethod
    def revoke_offer_sponsorship(
        cls,
        seller_id: str,
        offer_id: int,
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        """Create a **revoke sponsorship** operation for an offer.

        :param seller_id: The account ID which created the offer.
        :param offer_id: The offer ID.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for an offer.
        """
        offer = Offer(seller_id=seller_id, offer_id=offer_id)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.OFFER,
            account_id=None,
            trustline=None,
            offer=offer,
            data=None,
            claimable_balance_id=None,
            signer=None,
            liquidity_pool_id=None,
            source=source,
        )

    @classmethod
    def revoke_data_sponsorship(
        cls,
        account_id: str,
        data_name: str,
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        """Create a **revoke sponsorship** operation for a data entry.

        :param account_id: The account ID which owns the data entry.
        :param data_name: The name of the data entry
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for a data entry.
        """
        data = Data(account_id=account_id, data_name=data_name)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.DATA,
            account_id=None,
            trustline=None,
            offer=None,
            data=data,
            claimable_balance_id=None,
            signer=None,
            liquidity_pool_id=None,
            source=source,
        )

    @classmethod
    def revoke_claimable_balance_sponsorship(
        cls,
        claimable_balance_id: str,
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        """Create a **revoke sponsorship** operation for a claimable balance.

        :param claimable_balance_id: The sponsored claimable balance ID.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for a claimable balance.
        """
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.CLAIMABLE_BALANCE,
            account_id=None,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=claimable_balance_id,
            signer=None,
            liquidity_pool_id=None,
            source=source,
        )

    @classmethod
    def revoke_signer_sponsorship(
        cls,
        account_id: str,
        signer_key: SignerKey,
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        """Create a **revoke sponsorship** operation for a signer.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The signer whose sponsorship is being removed.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for a signer.
        """
        signer = Signer(account_id=account_id, signer_key=signer_key)
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.SIGNER,
            account_id=None,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=signer,
            liquidity_pool_id=None,
            source=source,
        )

    @classmethod
    def revoke_liquidity_pool_sponsorship(
        cls, liquidity_pool_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ):
        """Create a **revoke sponsorship** operation for a liquidity pool.

        :param liquidity_pool_id: The sponsored liquidity pool ID in hex string.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: A "revoke sponsorship" operation for a liquidity pool.
        """
        return cls(
            revoke_sponsorship_type=RevokeSponsorshipType.LIQUIDITY_POOL,
            account_id=None,
            trustline=None,
            offer=None,
            data=None,
            claimable_balance_id=None,
            signer=None,
            liquidity_pool_id=liquidity_pool_id,
            source=source,
        )

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        if self.revoke_sponsorship_type == RevokeSponsorshipType.ACCOUNT:
            assert self.account_id is not None
            account = stellar_xdr.LedgerKeyAccount(
                Keypair.from_public_key(self.account_id).xdr_account_id()
            )
            ledger_key = stellar_xdr.LedgerKey(
                stellar_xdr.LedgerEntryType.ACCOUNT, account=account
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY,
                ledger_key=ledger_key,
            )
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.TRUSTLINE:
            assert self.trustline is not None
            trust_line = stellar_xdr.LedgerKeyTrustLine(
                Keypair.from_public_key(self.trustline.account_id).xdr_account_id(),
                self.trustline.asset.to_trust_line_asset_xdr_object(),
            )
            ledger_key = stellar_xdr.LedgerKey(
                stellar_xdr.LedgerEntryType.TRUSTLINE, trust_line=trust_line
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY,
                ledger_key=ledger_key,
            )
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.OFFER:
            assert self.offer is not None
            offer = stellar_xdr.LedgerKeyOffer(
                Keypair.from_public_key(self.offer.seller_id).xdr_account_id(),
                stellar_xdr.Int64(self.offer.offer_id),
            )
            ledger_key = stellar_xdr.LedgerKey(
                stellar_xdr.LedgerEntryType.OFFER, offer=offer
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY,
                ledger_key=ledger_key,
            )
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.DATA:
            assert self.data is not None
            data = stellar_xdr.LedgerKeyData(
                Keypair.from_public_key(self.data.account_id).xdr_account_id(),
                stellar_xdr.String64(self.data.data_name.encode()),
            )
            ledger_key = stellar_xdr.LedgerKey(
                stellar_xdr.LedgerEntryType.DATA, data=data
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY,
                ledger_key=ledger_key,
            )
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.CLAIMABLE_BALANCE:
            assert self.claimable_balance_id is not None
            claimable_balance_bytes = binascii.unhexlify(self.claimable_balance_id)
            balance_id = stellar_xdr.ClaimableBalanceID.from_xdr_bytes(
                claimable_balance_bytes
            )
            claimable_balance = stellar_xdr.LedgerKeyClaimableBalance(
                balance_id=balance_id
            )
            ledger_key = stellar_xdr.LedgerKey(
                stellar_xdr.LedgerEntryType.CLAIMABLE_BALANCE,
                claimable_balance=claimable_balance,
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY,
                ledger_key=ledger_key,
            )
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.SIGNER:
            assert self.signer is not None
            signer_key = stellar_xdr.RevokeSponsorshipOpSigner(
                Keypair.from_public_key(self.signer.account_id).xdr_account_id(),
                self.signer.signer_key.to_xdr_object(),
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_SIGNER,
                signer=signer_key,
            )
        elif self.revoke_sponsorship_type == RevokeSponsorshipType.LIQUIDITY_POOL:
            assert self.liquidity_pool_id is not None
            liquidity_pool_id_bytes = binascii.unhexlify(self.liquidity_pool_id)
            pool_id = stellar_xdr.PoolID.from_xdr_bytes(liquidity_pool_id_bytes)
            liquidity_pool = stellar_xdr.LedgerKeyLiquidityPool(pool_id)
            ledger_key = stellar_xdr.LedgerKey(
                stellar_xdr.LedgerEntryType.LIQUIDITY_POOL,
                liquidity_pool=liquidity_pool,
            )
            revoke_sponsorship_op = stellar_xdr.RevokeSponsorshipOp(
                stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY,
                ledger_key=ledger_key,
            )
        else:
            raise ValueError(
                f"{self.revoke_sponsorship_type} is not a valid RevokeSponsorshipType."
            )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, revoke_sponsorship_op=revoke_sponsorship_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "RevokeSponsorship":
        """Creates a :class:`RevokeSponsorship` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.revoke_sponsorship_op is not None
        op_type = xdr_object.body.revoke_sponsorship_op.type
        if op_type == stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_LEDGER_ENTRY:
            assert xdr_object.body.revoke_sponsorship_op is not None
            ledger_key = xdr_object.body.revoke_sponsorship_op.ledger_key
            assert ledger_key is not None
            assert ledger_key.type is not None
            ledger_key_type = ledger_key.type
            if ledger_key_type == stellar_xdr.LedgerEntryType.ACCOUNT:
                assert ledger_key.account is not None
                assert ledger_key.account.account_id.account_id.ed25519 is not None
                account_id = StrKey.encode_ed25519_public_key(
                    ledger_key.account.account_id.account_id.ed25519.uint256
                )
                op = cls.revoke_account_sponsorship(account_id, source)
            elif ledger_key_type == stellar_xdr.LedgerEntryType.TRUSTLINE:
                assert ledger_key.trust_line is not None
                assert ledger_key.trust_line.account_id is not None
                assert ledger_key.trust_line.account_id.account_id.ed25519 is not None
                account_id = StrKey.encode_ed25519_public_key(
                    ledger_key.trust_line.account_id.account_id.ed25519.uint256
                )
                if (
                    ledger_key.trust_line.asset.type
                    == stellar_xdr.AssetType.ASSET_TYPE_POOL_SHARE
                ):
                    asset: Union[Asset, LiquidityPoolId] = (
                        LiquidityPoolId.from_xdr_object(ledger_key.trust_line.asset)
                    )
                else:
                    asset = Asset.from_xdr_object(ledger_key.trust_line.asset)
                op = cls.revoke_trustline_sponsorship(account_id, asset, source)
            elif ledger_key_type == stellar_xdr.LedgerEntryType.OFFER:
                assert ledger_key.offer is not None
                assert ledger_key.offer.seller_id.account_id.ed25519 is not None
                seller_id = StrKey.encode_ed25519_public_key(
                    ledger_key.offer.seller_id.account_id.ed25519.uint256
                )
                offer_id = ledger_key.offer.offer_id.int64
                op = cls.revoke_offer_sponsorship(seller_id, offer_id, source)
            elif ledger_key_type == stellar_xdr.LedgerEntryType.DATA:
                assert ledger_key.data is not None
                assert ledger_key.data.account_id is not None
                assert ledger_key.data.account_id.account_id is not None
                assert ledger_key.data.account_id.account_id.ed25519 is not None
                account_id = StrKey.encode_ed25519_public_key(
                    ledger_key.data.account_id.account_id.ed25519.uint256
                )
                data_name = ledger_key.data.data_name.string64.decode()
                op = cls.revoke_data_sponsorship(account_id, data_name, source)
            elif ledger_key_type == stellar_xdr.LedgerEntryType.CLAIMABLE_BALANCE:
                assert ledger_key.claimable_balance is not None
                balance_id_bytes = base64.b64decode(
                    ledger_key.claimable_balance.balance_id.to_xdr()
                )
                balance_id = binascii.hexlify(balance_id_bytes).decode()
                op = cls.revoke_claimable_balance_sponsorship(balance_id, source)
            elif ledger_key_type == stellar_xdr.LedgerEntryType.LIQUIDITY_POOL:
                assert ledger_key.liquidity_pool is not None
                assert ledger_key.liquidity_pool.liquidity_pool_id is not None
                assert ledger_key.liquidity_pool.liquidity_pool_id.pool_id is not None
                liquidity_pool_id = (
                    ledger_key.liquidity_pool.liquidity_pool_id.pool_id.hash.hex()
                )
                op = cls.revoke_liquidity_pool_sponsorship(liquidity_pool_id, source)
            else:
                raise ValueError(f"{ledger_key_type} is an unsupported LedgerKey type.")
        elif op_type == stellar_xdr.RevokeSponsorshipType.REVOKE_SPONSORSHIP_SIGNER:
            assert xdr_object.body.revoke_sponsorship_op.signer is not None
            assert (
                xdr_object.body.revoke_sponsorship_op.signer.account_id.account_id.ed25519
                is not None
            )
            account_id = StrKey.encode_ed25519_public_key(
                xdr_object.body.revoke_sponsorship_op.signer.account_id.account_id.ed25519.uint256
            )
            signer_key = SignerKey.from_xdr_object(
                xdr_object.body.revoke_sponsorship_op.signer.signer_key
            )
            op = cls.revoke_signer_sponsorship(account_id, signer_key, source)
        else:
            raise ValueError(f"{op_type} is an unsupported RevokeSponsorship type.")
        return op

    def __hash__(self):
        return hash(
            (
                self.revoke_sponsorship_type,
                self.account_id,
                self.trustline,
                self.offer,
                self.data,
                self.claimable_balance_id,
                self.liquidity_pool_id,
                self.signer,
                self.source,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.revoke_sponsorship_type == other.revoke_sponsorship_type
            and self.account_id == other.account_id
            and self.trustline == other.trustline
            and self.offer == other.offer
            and self.data == other.data
            and self.claimable_balance_id == other.claimable_balance_id
            and self.liquidity_pool_id == other.liquidity_pool_id
            and self.signer == other.signer
            and self.source == other.source
        )

    def __repr__(self):
        return (
            f"<RevokeSponsorship [revoke_sponsorship_type={self.revoke_sponsorship_type}, "
            f"account_id={self.account_id}, "
            f"trustline={self.trustline}, "
            f"offer={self.offer}, "
            f"data={self.data}, "
            f"claimable_balance_id={self.claimable_balance_id}, "
            f"liquidity_pool_id={self.liquidity_pool_id}, "
            f"signer={self.signer}, "
            f"source={self.source}]>"
        )



from typing import Any, Dict, Generator

import requests
from requests import RequestException
from urllib3.exceptions import NewConnectionError

from ..__version__ import __version__
from ..client.response import Response
from .base_sync_client import BaseSyncClient

USER_AGENT = f"py-stellar-base/{__version__}/SimpleRequestsClient"
HEADERS = {
    "X-Client-Name": "py-stellar-base",
    "X-Client-Version": __version__,
    "User-Agent": USER_AGENT,
}

__all__ = ["SimpleRequestsClient"]


class SimpleRequestsClient(BaseSyncClient):
    """The :class:`SimpleRequestsClient` object is a synchronous http client,
    which represents the interface for making requests to a server instance.

    **This client is to guide you in writing a client that suits your needs.
    I don't recommend that you actually use it.**
    """

    def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            resp = requests.get(url=url, params=params, headers=HEADERS)
        except (RequestException, NewConnectionError) as err:
            raise ConnectionError(err)
        return Response(
            status_code=resp.status_code,
            text=resp.text,
            headers=dict(resp.headers),
            url=resp.url,
        )

    def post(
        self, url: str, data: Dict[str, str] = None, json_data: Dict[str, Any] = None
    ) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :param json_data: the json data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            resp = requests.post(url=url, data=data, json=json_data, headers=HEADERS)
        except (RequestException, NewConnectionError) as err:
            raise ConnectionError(err)
        return Response(
            status_code=resp.status_code,
            text=resp.text,
            headers=dict(resp.headers),
            url=resp.url,
        )

    def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """
        **Not Implemented**

        :param url: the request url
        :param params: the request params
        :return: None
        """
        raise NotImplementedError  # pragma: no cover

    def close(self):
        pass  # pragma: no cover

    def __repr__(self):
        return "<SimpleRequestsClient>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Opaque
from .uint256 import Uint256

__all__ = ["SignerKeyEd25519SignedPayload"]


class SignerKeyEd25519SignedPayload:
    """
    XDR Source Code::

        struct
            {
                /* Public key that must sign the payload. */
                uint256 ed25519;
                /* Payload to be raw signed by ed25519. */
                opaque payload<64>;
            }
    """

    def __init__(
        self,
        ed25519: Uint256,
        payload: bytes,
    ) -> None:
        self.ed25519 = ed25519
        self.payload = payload

    def pack(self, packer: Packer) -> None:
        self.ed25519.pack(packer)
        Opaque(self.payload, 64, False).pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SignerKeyEd25519SignedPayload:
        ed25519 = Uint256.unpack(unpacker)
        payload = Opaque.unpack(unpacker, 64, False)
        return cls(
            ed25519=ed25519,
            payload=payload,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SignerKeyEd25519SignedPayload:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SignerKeyEd25519SignedPayload:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ed25519,
                self.payload,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ed25519 == other.ed25519 and self.payload == other.payload

    def __repr__(self):
        out = [
            f"ed25519={self.ed25519}",
            f"payload={self.payload}",
        ]
        return f"<SignerKeyEd25519SignedPayload [{', '.join(out)}]>"
